# transaction.py - simple journaling scheme for mercurial
#
# This transaction scheme is intended to gracefully handle program
# errors and interruptions. More serious failures like system crashes
# can be recovered with an fsck-like tool. As the whole repository is
# effectively log-structured, this should amount to simply truncating
# anything that isn't referenced in the changelog.
#
# Copyright 2005, 2006 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import abc

from typing import (
    Callable,
    Collection,
    ContextManager,
    Protocol,
)

from ._basetypes import (
    CallbackCategoryT,
    HgPathT,
    VfsKeyT,
)

JournalEntryT = tuple[HgPathT, int]


class ITransaction(ContextManager, Protocol):
    @property
    @abc.abstractmethod
    def finalized(self) -> bool:
        ...

    @abc.abstractmethod
    def startgroup(self) -> None:
        """delay registration of file entry

        This is used by strip to delay vision of strip offset. The transaction
        sees either none or all of the strip actions to be done."""

    @abc.abstractmethod
    def endgroup(self) -> None:
        """apply delayed registration of file entry.

        This is used by strip to delay vision of strip offset. The transaction
        sees either none or all of the strip actions to be done."""

    @abc.abstractmethod
    def add(self, file: HgPathT, offset: int) -> None:
        """record the state of an append-only file before update"""

    @abc.abstractmethod
    def addbackup(
        self,
        file: HgPathT,
        hardlink: bool = True,
        location: VfsKeyT = b'',
        for_offset: bool | int = False,
    ) -> None:
        """Adds a backup of the file to the transaction

        Calling addbackup() creates a hardlink backup of the specified file
        that is used to recover the file in the event of the transaction
        aborting.

        * `file`: the file path, relative to .hg/store
        * `hardlink`: use a hardlink to quickly create the backup

        If `for_offset` is set, we expect a offset for this file to have been
        previously recorded
        """

    @abc.abstractmethod
    def registertmp(self, tmpfile: HgPathT, location: VfsKeyT = b'') -> None:
        """register a temporary transaction file

        Such files will be deleted when the transaction exits (on both
        failure and success).
        """

    @abc.abstractmethod
    def addfilegenerator(
        self,
        genid: bytes,
        filenames: Collection[HgPathT],
        genfunc: Callable,
        order: int = 0,
        location: VfsKeyT = b'',
        post_finalize: bool = False,
    ) -> None:
        """add a function to generates some files at transaction commit

        The `genfunc` argument is a function capable of generating proper
        content of each entry in the `filename` tuple.

        At transaction close time, `genfunc` will be called with one file
        object argument per entries in `filenames`.

        The transaction itself is responsible for the backup, creation and
        final write of such file.

        The `genid` argument is used to ensure the same set of file is only
        generated once. Call to `addfilegenerator` for a `genid` already
        present will overwrite the old entry.

        The `order` argument may be used to control the order in which multiple
        generator will be executed.

        The `location` arguments may be used to indicate the files are located
        outside of the the standard directory for transaction. It should match
        one of the key of the `transaction.vfsmap` dictionary.

        The `post_finalize` argument can be set to `True` for file generation
        that must be run after the transaction has been finalized.
        """

    @abc.abstractmethod
    def removefilegenerator(self, genid: bytes) -> None:
        """reverse of addfilegenerator, remove a file generator function"""

    @abc.abstractmethod
    def findoffset(self, file: HgPathT) -> int | None:
        ...

    @abc.abstractmethod
    def readjournal(self) -> list[JournalEntryT]:
        ...

    @abc.abstractmethod
    def replace(self, file: HgPathT, offset: int) -> None:
        """
        replace can only replace already committed entries
        that are not pending in the queue
        """

    @abc.abstractmethod
    def nest(self, name: bytes = b'<unnamed>') -> ITransaction:
        ...

    @abc.abstractmethod
    def release(self) -> None:
        ...

    @abc.abstractmethod
    def running(self) -> bool:
        ...

    @abc.abstractmethod
    def addpending(
        self,
        category: CallbackCategoryT,
        callback: Callable[[ITransaction], None],
    ) -> None:
        """add a callback to be called when the transaction is pending

        The transaction will be given as callback's first argument.

        Category is a unique identifier to allow overwriting an old callback
        with a newer callback.
        """

    @abc.abstractmethod
    def writepending(self) -> bool:
        """write pending file to temporary version

        This is used to allow hooks to view a transaction before commit"""

    @abc.abstractmethod
    def hasfinalize(self, category: CallbackCategoryT) -> bool:
        """check is a callback already exist for a category"""

    @abc.abstractmethod
    def addfinalize(
        self,
        category: CallbackCategoryT,
        callback: Callable[[ITransaction], None],
    ) -> None:
        """add a callback to be called when the transaction is closed

        The transaction will be given as callback's first argument.

        Category is a unique identifier to allow overwriting old callbacks with
        newer callbacks.
        """

    @abc.abstractmethod
    def addpostclose(
        self,
        category: CallbackCategoryT,
        callback: Callable[[ITransaction], None],
    ) -> None:
        """add or replace a callback to be called after the transaction closed

        The transaction will be given as callback's first argument.

        Category is a unique identifier to allow overwriting an old callback
        with a newer callback.
        """

    @abc.abstractmethod
    def getpostclose(
        self, category: CallbackCategoryT
    ) -> Callable[[ITransaction], None] | None:
        """return a postclose callback added before, or None"""

    @abc.abstractmethod
    def addabort(
        self,
        category: CallbackCategoryT,
        callback: Callable[[ITransaction], None],
    ) -> None:
        """add a callback to be called when the transaction is aborted.

        The transaction will be given as the first argument to the callback.

        Category is a unique identifier to allow overwriting an old callback
        with a newer callback.
        """

    @abc.abstractmethod
    def addvalidator(
        self,
        category: CallbackCategoryT,
        callback: Callable[[ITransaction], None],
    ) -> None:
        """adds a callback to be called when validating the transaction.

        The transaction will be given as the first argument to the callback.

        callback should raise exception if to abort transaction"""

    @abc.abstractmethod
    def close(self) -> None:
        '''commit the transaction'''

    @abc.abstractmethod
    def abort(self) -> None:
        """abort the transaction (generally called on error, or when the
        transaction is not explicitly committed before going out of
        scope)"""

    @abc.abstractmethod
    def add_journal(self, vfs_id: VfsKeyT, path: HgPathT) -> None:
        ...



# utility for color output for Mercurial commands
#
# Copyright (C) 2007 Kevin Christen <kevin.christen@gmail.com> and other
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import re

from .i18n import _

from . import (
    encoding,
    pycompat,
)

from .utils import stringutil

try:
    import curses

    # Mapping from effect name to terminfo attribute name (or raw code) or
    # color number.  This will also force-load the curses module.
    _baseterminfoparams = {
        b'none': (True, b'sgr0', b''),
        b'standout': (True, b'smso', b''),
        b'underline': (True, b'smul', b''),
        b'reverse': (True, b'rev', b''),
        b'inverse': (True, b'rev', b''),
        b'blink': (True, b'blink', b''),
        b'dim': (True, b'dim', b''),
        b'bold': (True, b'bold', b''),
        b'invisible': (True, b'invis', b''),
        b'italic': (True, b'sitm', b''),
        b'black': (False, curses.COLOR_BLACK, b''),
        b'red': (False, curses.COLOR_RED, b''),
        b'green': (False, curses.COLOR_GREEN, b''),
        b'yellow': (False, curses.COLOR_YELLOW, b''),
        b'blue': (False, curses.COLOR_BLUE, b''),
        b'magenta': (False, curses.COLOR_MAGENTA, b''),
        b'cyan': (False, curses.COLOR_CYAN, b''),
        b'white': (False, curses.COLOR_WHITE, b''),
    }
except (ImportError, AttributeError):
    curses = None
    _baseterminfoparams = {}

# start and stop parameters for effects
_effects = {
    b'none': 0,
    b'black': 30,
    b'red': 31,
    b'green': 32,
    b'yellow': 33,
    b'blue': 34,
    b'magenta': 35,
    b'cyan': 36,
    b'white': 37,
    b'bold': 1,
    b'italic': 3,
    b'underline': 4,
    b'inverse': 7,
    b'dim': 2,
    b'black_background': 40,
    b'red_background': 41,
    b'green_background': 42,
    b'yellow_background': 43,
    b'blue_background': 44,
    b'purple_background': 45,
    b'cyan_background': 46,
    b'white_background': 47,
}

_defaultstyles = {
    b'grep.match': b'red bold',
    b'grep.linenumber': b'green',
    b'grep.rev': b'blue',
    b'grep.sep': b'cyan',
    b'grep.filename': b'magenta',
    b'grep.user': b'magenta',
    b'grep.date': b'magenta',
    b'grep.inserted': b'green bold',
    b'grep.deleted': b'red bold',
    b'bookmarks.active': b'green',
    b'branches.active': b'none',
    b'branches.closed': b'black bold',
    b'branches.current': b'green',
    b'branches.inactive': b'none',
    b'diff.changed': b'white',
    b'diff.deleted': b'red',
    b'diff.deleted.changed': b'red bold underline',
    b'diff.deleted.unchanged': b'red',
    b'diff.diffline': b'bold',
    b'diff.extended': b'cyan bold',
    b'diff.file_a': b'red bold',
    b'diff.file_b': b'green bold',
    b'diff.hunk': b'magenta',
    b'diff.inserted': b'green',
    b'diff.inserted.changed': b'green bold underline',
    b'diff.inserted.unchanged': b'green',
    b'diff.tab': b'',
    b'diff.trailingwhitespace': b'bold red_background',
    b'changeset.public': b'',
    b'changeset.draft': b'',
    b'changeset.secret': b'',
    b'diffstat.deleted': b'red',
    b'diffstat.inserted': b'green',
    b'formatvariant.name.mismatchconfig': b'red',
    b'formatvariant.name.mismatchdefault': b'yellow',
    b'formatvariant.name.uptodate': b'green',
    b'formatvariant.repo.mismatchconfig': b'red',
    b'formatvariant.repo.mismatchdefault': b'yellow',
    b'formatvariant.repo.uptodate': b'green',
    b'formatvariant.config.special': b'yellow',
    b'formatvariant.config.default': b'green',
    b'formatvariant.default': b'',
    b'histedit.remaining': b'red bold',
    b'ui.addremove.added': b'green',
    b'ui.addremove.removed': b'red',
    b'ui.error': b'red',
    b'ui.prompt': b'yellow',
    b'log.changeset': b'yellow',
    b'patchbomb.finalsummary': b'',
    b'patchbomb.from': b'magenta',
    b'patchbomb.to': b'cyan',
    b'patchbomb.subject': b'green',
    b'patchbomb.diffstats': b'',
    b'rebase.rebased': b'blue',
    b'rebase.remaining': b'red bold',
    b'resolve.resolved': b'green bold',
    b'resolve.unresolved': b'red bold',
    b'shelve.age': b'cyan',
    b'shelve.newest': b'green bold',
    b'shelve.name': b'blue bold',
    b'status.added': b'green bold',
    b'status.clean': b'none',
    b'status.copied': b'none',
    b'status.deleted': b'cyan bold underline',
    b'status.ignored': b'black bold',
    b'status.modified': b'blue bold',
    b'status.removed': b'red bold',
    b'status.unknown': b'magenta bold underline',
    b'tags.normal': b'green',
    b'tags.local': b'black bold',
    b'upgrade-repo.requirement.preserved': b'cyan',
    b'upgrade-repo.requirement.added': b'green',
    b'upgrade-repo.requirement.removed': b'red',
}


def loadcolortable(ui, extname, colortable):
    _defaultstyles.update(colortable)


def _terminfosetup(ui, mode, formatted):
    '''Initialize terminfo data and the terminal if we're in terminfo mode.'''

    # If we failed to load curses, we go ahead and return.
    if curses is None:
        return
    # Otherwise, see what the config file says.
    if mode not in (b'auto', b'terminfo'):
        return
    ui._terminfoparams.update(_baseterminfoparams)

    for key, val in ui.configitems(b'color'):
        if key.startswith(b'color.'):
            newval = (False, int(val), b'')
            ui._terminfoparams[key[6:]] = newval
        elif key.startswith(b'terminfo.'):
            newval = (True, b'', val.replace(b'\\E', b'\x1b'))
            ui._terminfoparams[key[9:]] = newval
    try:
        curses.setupterm()
    except curses.error:
        ui._terminfoparams.clear()
        return

    for key, (b, e, c) in ui._terminfoparams.copy().items():
        if not b:
            continue
        if not c and not curses.tigetstr(pycompat.sysstr(e)):
            # Most terminals don't support dim, invis, etc, so don't be
            # noisy and use ui.debug().
            ui.debug(b"no terminfo entry for %s\n" % e)
            del ui._terminfoparams[key]
    if not curses.tigetstr('setaf') or not curses.tigetstr('setab'):
        # Only warn about missing terminfo entries if we explicitly asked for
        # terminfo mode and we're in a formatted terminal.
        if mode == b"terminfo" and formatted:
            ui.warn(
                _(
                    b"no terminfo entry for setab/setaf: reverting to "
                    b"ECMA-48 color\n"
                )
            )
        ui._terminfoparams.clear()


def setup(ui):
    """configure color on a ui

    That function both set the colormode for the ui object and read
    the configuration looking for custom colors and effect definitions."""
    mode = _modesetup(ui)
    ui._colormode = mode
    if mode and mode != b'debug':
        configstyles(ui)


def _modesetup(ui):
    if ui.plain(b'color'):
        return None
    config = ui.config(b'ui', b'color')
    if config == b'debug':
        return b'debug'

    auto = config == b'auto'
    always = False
    if not auto and stringutil.parsebool(config):
        # We want the config to behave like a boolean, "on" is actually auto,
        # but "always" value is treated as a special case to reduce confusion.
        if (
            ui.configsource(b'ui', b'color') == b'--color'
            or config == b'always'
        ):
            always = True
        else:
            auto = True

    if not always and not auto:
        return None

    formatted = always or (
        encoding.environ.get(b'TERM') != b'dumb' and ui.formatted()
    )

    mode = ui.config(b'color', b'mode')

    # If pager is active, color.pagermode overrides color.mode.
    if getattr(ui, 'pageractive', False):
        mode = ui.config(b'color', b'pagermode', mode)

    realmode = mode
    if pycompat.iswindows:
        from . import win32

        if mode == b'auto':
            # Since "ansi" could result in terminal gibberish, we error on the
            # side of selecting "win32". However, if w32effects is not defined,
            # we almost certainly don't support "win32", so don't even try.
            # w32effects is not populated when stdout is redirected, so checking
            # it first avoids win32 calls in a state known to error out.
            if not w32effects or win32.enablevtmode():
                realmode = b'ansi'
            else:
                realmode = b'win32'
        # An empty w32effects is a clue that stdout is redirected, and thus
        # cannot enable VT mode.
        elif mode == b'ansi' and w32effects:
            win32.enablevtmode()
    elif mode == b'auto':
        realmode = b'ansi'

    def modewarn():
        # only warn if color.mode was explicitly set and we're in
        # a formatted terminal
        if mode == realmode and formatted:
            ui.warn(_(b'warning: failed to set color mode to %s\n') % mode)

    if realmode == b'win32':
        ui._terminfoparams.clear()
        if not w32effects:
            modewarn()
            return None
    elif realmode == b'ansi':
        ui._terminfoparams.clear()
    elif realmode == b'terminfo':
        _terminfosetup(ui, mode, formatted)
        if not ui._terminfoparams:
            ## FIXME Shouldn't we return None in this case too?
            modewarn()
            realmode = b'ansi'
    else:
        return None

    if always or (auto and formatted):
        return realmode
    return None


def configstyles(ui):
    ui._styles.update(_defaultstyles)
    for status, cfgeffects in ui.configitems(b'color'):
        if b'.' not in status or status.startswith((b'color.', b'terminfo.')):
            continue
        cfgeffects = ui.configlist(b'color', status)
        if cfgeffects:
            good = []
            for e in cfgeffects:
                if valideffect(ui, e):
                    good.append(e)
                else:
                    ui.warn(
                        _(
                            b"ignoring unknown color/effect %s "
                            b"(configured in color.%s)\n"
                        )
                        % (stringutil.pprint(e), status)
                    )
            ui._styles[status] = b' '.join(good)


def _activeeffects(ui):
    '''Return the effects map for the color mode set on the ui.'''
    if ui._colormode == b'win32':
        return w32effects
    elif ui._colormode is not None:
        return _effects
    return {}


def valideffect(ui, effect):
    """Determine if the effect is valid or not."""
    return (not ui._terminfoparams and effect in _activeeffects(ui)) or (
        effect in ui._terminfoparams or effect[:-11] in ui._terminfoparams
    )


def _effect_str(ui, effect):
    '''Helper function for render_effects().'''

    bg = False
    if effect.endswith(b'_background'):
        bg = True
        effect = effect[:-11]
    try:
        attr, val, termcode = ui._terminfoparams[effect]
    except KeyError:
        return b''
    if attr:
        if termcode:
            return termcode
        else:
            return curses.tigetstr(pycompat.sysstr(val))
    elif bg:
        return curses.tparm(curses.tigetstr('setab'), val)
    else:
        return curses.tparm(curses.tigetstr('setaf'), val)


def _mergeeffects(text, start, stop):
    """Insert start sequence at every occurrence of stop sequence

    >>> s = _mergeeffects(b'cyan', b'[C]', b'|')
    >>> s = _mergeeffects(s + b'yellow', b'[Y]', b'|')
    >>> s = _mergeeffects(b'ma' + s + b'genta', b'[M]', b'|')
    >>> s = _mergeeffects(b'red' + s, b'[R]', b'|')
    >>> s
    '[R]red[M]ma[Y][C]cyan|[R][M][Y]yellow|[R][M]genta|'
    """
    parts = []
    for t in text.split(stop):
        if not t:
            continue
        parts.extend([start, t, stop])
    return b''.join(parts)


def _render_effects(ui, text, effects):
    """Wrap text in commands to turn on each effect."""
    if not text:
        return text
    if ui._terminfoparams:
        start = b''.join(
            _effect_str(ui, effect) for effect in [b'none'] + effects.split()
        )
        stop = _effect_str(ui, b'none')
    else:
        activeeffects = _activeeffects(ui)
        start = [
            pycompat.bytestr(activeeffects[e])
            for e in [b'none'] + effects.split()
        ]
        start = b'\033[' + b';'.join(start) + b'm'
        stop = b'\033[' + pycompat.bytestr(activeeffects[b'none']) + b'm'
        if ui._readlineprompt:
            start = b'\001' + start + b'\001'
            stop = b'\002' + stop + b'\002'
    return _mergeeffects(text, start, stop)


_ansieffectre = re.compile(br'\x1b\[[0-9;]*m')


def stripeffects(text):
    """Strip ANSI control codes which could be inserted by colorlabel()"""
    return _ansieffectre.sub(b'', text)


def colorlabel(ui, msg, label):
    """add color control code according to the mode"""
    if ui._colormode == b'debug':
        if label and msg:
            if msg.endswith(b'\n'):
                msg = b"[%s|%s]\n" % (label, msg[:-1])
            else:
                msg = b"[%s|%s]" % (label, msg)
    elif ui._colormode is not None:
        effects = []
        for l in label.split():
            s = ui._styles.get(l, b'')
            if s:
                effects.append(s)
            elif valideffect(ui, l):
                effects.append(l)
        effects = b' '.join(effects)
        if effects:
            msg = b'\n'.join(
                [
                    _render_effects(ui, line, effects)
                    for line in msg.split(b'\n')
                ]
            )
    return msg


w32effects = None
if pycompat.iswindows:
    import ctypes

    _kernel32 = ctypes.windll.kernel32  # pytype: disable=module-attr

    _WORD = ctypes.c_ushort

    _INVALID_HANDLE_VALUE = -1

    class _COORD(ctypes.Structure):
        _fields_ = [('X', ctypes.c_short), ('Y', ctypes.c_short)]

    class _SMALL_RECT(ctypes.Structure):
        _fields_ = [
            ('Left', ctypes.c_short),
            ('Top', ctypes.c_short),
            ('Right', ctypes.c_short),
            ('Bottom', ctypes.c_short),
        ]

    class _CONSOLE_SCREEN_BUFFER_INFO(ctypes.Structure):
        _fields_ = [
            ('dwSize', _COORD),
            ('dwCursorPosition', _COORD),
            ('wAttributes', _WORD),
            ('srWindow', _SMALL_RECT),
            ('dwMaximumWindowSize', _COORD),
        ]

    _STD_OUTPUT_HANDLE = 0xFFFFFFF5  # (DWORD)-11
    _STD_ERROR_HANDLE = 0xFFFFFFF4  # (DWORD)-12

    _FOREGROUND_BLUE = 0x0001
    _FOREGROUND_GREEN = 0x0002
    _FOREGROUND_RED = 0x0004
    _FOREGROUND_INTENSITY = 0x0008

    _BACKGROUND_BLUE = 0x0010
    _BACKGROUND_GREEN = 0x0020
    _BACKGROUND_RED = 0x0040
    _BACKGROUND_INTENSITY = 0x0080

    _COMMON_LVB_REVERSE_VIDEO = 0x4000
    _COMMON_LVB_UNDERSCORE = 0x8000

    # http://msdn.microsoft.com/en-us/library/ms682088%28VS.85%29.aspx
    w32effects = {
        b'none': -1,
        b'black': 0,
        b'red': _FOREGROUND_RED,
        b'green': _FOREGROUND_GREEN,
        b'yellow': _FOREGROUND_RED | _FOREGROUND_GREEN,
        b'blue': _FOREGROUND_BLUE,
        b'magenta': _FOREGROUND_BLUE | _FOREGROUND_RED,
        b'cyan': _FOREGROUND_BLUE | _FOREGROUND_GREEN,
        b'white': _FOREGROUND_RED | _FOREGROUND_GREEN | _FOREGROUND_BLUE,
        b'bold': _FOREGROUND_INTENSITY,
        b'black_background': 0x100,  # unused value > 0x0f
        b'red_background': _BACKGROUND_RED,
        b'green_background': _BACKGROUND_GREEN,
        b'yellow_background': _BACKGROUND_RED | _BACKGROUND_GREEN,
        b'blue_background': _BACKGROUND_BLUE,
        b'purple_background': _BACKGROUND_BLUE | _BACKGROUND_RED,
        b'cyan_background': _BACKGROUND_BLUE | _BACKGROUND_GREEN,
        b'white_background': (
            _BACKGROUND_RED | _BACKGROUND_GREEN | _BACKGROUND_BLUE
        ),
        b'bold_background': _BACKGROUND_INTENSITY,
        b'underline': _COMMON_LVB_UNDERSCORE,  # double-byte charsets only
        b'inverse': _COMMON_LVB_REVERSE_VIDEO,  # double-byte charsets only
    }

    passthrough = {
        _FOREGROUND_INTENSITY,
        _BACKGROUND_INTENSITY,
        _COMMON_LVB_UNDERSCORE,
        _COMMON_LVB_REVERSE_VIDEO,
    }

    stdout = _kernel32.GetStdHandle(
        _STD_OUTPUT_HANDLE
    )  # don't close the handle returned
    if stdout is None or stdout == _INVALID_HANDLE_VALUE:
        w32effects = None
    else:
        csbi = _CONSOLE_SCREEN_BUFFER_INFO()
        if not _kernel32.GetConsoleScreenBufferInfo(stdout, ctypes.byref(csbi)):
            # stdout may not support GetConsoleScreenBufferInfo()
            # when called from subprocess or redirected
            w32effects = None
        else:
            origattr = csbi.wAttributes
            ansire = re.compile(
                br'\001?\033\[([^m]*)m\002?([^\033]*)(.*)',
                re.MULTILINE | re.DOTALL,
            )

    def win32print(ui, writefunc, text, **opts):
        label = opts.get('label', b'')
        attr = origattr

        def mapcolor(val, attr):
            if val == -1:
                return origattr
            elif val in passthrough:
                return attr | val
            elif val > 0x0F:
                return (val & 0x70) | (attr & 0x8F)
            else:
                return (val & 0x07) | (attr & 0xF8)

        # determine console attributes based on labels
        for l in label.split():
            style = ui._styles.get(l, b'')
            for effect in style.split():
                try:
                    attr = mapcolor(w32effects[effect], attr)
                except KeyError:
                    # w32effects could not have certain attributes so we skip
                    # them if not found
                    pass
        # hack to ensure regexp finds data
        if not text.startswith(b'\033['):
            text = b'\033[m' + text

        # Look for ANSI-like codes embedded in text
        m = re.match(ansire, text)

        try:
            while m:
                for sattr in m.group(1).split(b';'):
                    if sattr:
                        attr = mapcolor(int(sattr), attr)
                ui.flush()
                _kernel32.SetConsoleTextAttribute(stdout, attr)
                writefunc(m.group(2))
                m = re.match(ansire, m.group(3))
        finally:
            # Explicitly reset original attributes
            ui.flush()
            _kernel32.SetConsoleTextAttribute(stdout, origattr)



# streamclone.py - producing and consuming streaming repository data
#
# Copyright 2015 Gregory Szorc <gregory.szorc@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import collections
import contextlib
import errno
import os
import struct
import threading

from typing import (
    Callable,
    Iterable,
    Iterator,
    Optional,
)

from .i18n import _
from .interfaces import repository
from . import (
    bookmarks,
    bundle2 as bundle2mod,
    cacheutil,
    error,
    narrowspec,
    phases,
    pycompat,
    requirements as requirementsmod,
    scmutil,
    store,
    transaction,
    util,
)
from .revlogutils import (
    nodemap,
)


# Number arbitrarily picked, feel free to change them (but the LOW one)
#
# update the configuration documentation if you touch this.
DEFAULT_NUM_WRITER = {
    scmutil.RESOURCE_LOW: 1,
    scmutil.RESOURCE_MEDIUM: 4,
    scmutil.RESOURCE_HIGH: 8,
}


# Number arbitrarily picked, feel free to adjust them. Do update the
# documentation if you do so
DEFAULT_MEMORY_TARGET = {
    scmutil.RESOURCE_LOW: 50 * (2**20),  # 100 MB
    scmutil.RESOURCE_MEDIUM: 500 * 2**20,  # 500 MB
    scmutil.RESOURCE_HIGH: 2 * 2**30,  #   2 GB
}


def new_stream_clone_requirements(
    default_requirements: Iterable[bytes],
    streamed_requirements: Iterable[bytes],
) -> set[bytes]:
    """determine the final set of requirement for a new stream clone

    this method combine the "default" requirements that a new repository would
    use with the constaint we get from the stream clone content. We keep local
    configuration choice when possible.
    """
    requirements = set(default_requirements)
    requirements &= requirementsmod.STREAM_IGNORABLE_REQUIREMENTS | {
        requirementsmod.NARROW_REQUIREMENT
    }
    requirements.update(streamed_requirements)
    return requirements


def streamed_requirements(repo) -> set[bytes]:
    """the set of requirement the new clone will have to support

    This is used for advertising the stream options and to generate the actual
    stream content."""
    requiredformats = (
        repo.requirements - requirementsmod.STREAM_IGNORABLE_REQUIREMENTS
    )
    return requiredformats


def canperformstreamclone(pullop, bundle2: bool = False):
    """Whether it is possible to perform a streaming clone as part of pull.

    ``bundle2`` will cause the function to consider stream clone through
    bundle2 and only through bundle2.

    Returns a tuple of (supported, requirements). ``supported`` is True if
    streaming clone is supported and False otherwise. ``requirements`` is
    a set of repo requirements from the remote, or ``None`` if stream clone
    isn't supported.
    """
    repo = pullop.repo
    remote = pullop.remote

    # should we consider streaming clone at all ?
    streamrequested = pullop.streamclonerequested
    # If we don't have a preference, let the server decide for us. This
    # likely only comes into play in LANs.
    if streamrequested is None:
        # The server can advertise whether to prefer streaming clone.
        streamrequested = remote.capable(b'stream-preferred')
    if not streamrequested:
        return False, None

    # Streaming clone only works on an empty destination repository
    if len(repo):
        return False, None

    # Streaming clone only works if all data is being requested.
    if pullop.heads:
        return False, None

    bundle2supported = False
    if pullop.canusebundle2:
        local_caps = bundle2mod.getrepocaps(repo, role=b'client')
        local_supported = set(local_caps.get(b'stream', []))
        remote_supported = set(pullop.remotebundle2caps.get(b'stream', []))
        bundle2supported = bool(local_supported & remote_supported)
        # else
        # Server doesn't support bundle2 stream clone or doesn't support
        # the versions we support. Fall back and possibly allow legacy.

    # Ensures legacy code path uses available bundle2.
    if bundle2supported and not bundle2:
        return False, None
    # Ensures bundle2 doesn't try to do a stream clone if it isn't supported.
    elif bundle2 and not bundle2supported:
        return False, None

    # In order for stream clone to work, the client has to support all the
    # requirements advertised by the server.
    #
    # The server advertises its requirements via the "stream" and "streamreqs"
    # capability. "stream" (a value-less capability) is advertised if and only
    # if the only requirement is "revlogv1." Else, the "streamreqs" capability
    # is advertised and contains a comma-delimited list of requirements.
    requirements = set()
    if remote.capable(b'stream'):
        requirements.add(requirementsmod.REVLOGV1_REQUIREMENT)
    else:
        streamreqs = remote.capable(b'streamreqs')
        # This is weird and shouldn't happen with modern servers.
        if not streamreqs:
            pullop.repo.ui.warn(
                _(
                    b'warning: stream clone requested but server has them '
                    b'disabled\n'
                )
            )
            return False, None

        streamreqs = set(streamreqs.split(b','))
        # Server requires something we don't support. Bail.
        missingreqs = streamreqs - repo.supported
        if missingreqs:
            pullop.repo.ui.warn(
                _(
                    b'warning: stream clone requested but client is missing '
                    b'requirements: %s\n'
                )
                % b', '.join(sorted(missingreqs))
            )
            pullop.repo.ui.warn(
                _(
                    b'(see https://www.mercurial-scm.org/wiki/MissingRequirement '
                    b'for more information)\n'
                )
            )
            return False, None
        requirements = streamreqs

    return True, requirements


def maybeperformlegacystreamclone(pullop) -> None:
    """Possibly perform a legacy stream clone operation.

    Legacy stream clones are performed as part of pull but before all other
    operations.

    A legacy stream clone will not be performed if a bundle2 stream clone is
    supported.
    """
    from . import localrepo

    supported, requirements = canperformstreamclone(pullop)

    if not supported:
        return

    repo = pullop.repo
    remote = pullop.remote

    # Save remote branchmap. We will use it later to speed up branchcache
    # creation.
    rbranchmap = None
    if remote.capable(b'branchmap'):
        with remote.commandexecutor() as e:
            rbranchmap = e.callcommand(b'branchmap', {}).result()

    repo.ui.status(_(b'streaming all changes\n'))

    with remote.commandexecutor() as e:
        fp = e.callcommand(b'stream_out', {}).result()

    # TODO strictly speaking, this code should all be inside the context
    # manager because the context manager is supposed to ensure all wire state
    # is flushed when exiting. But the legacy peers don't do this, so it
    # doesn't matter.
    l = fp.readline()
    try:
        resp = int(l)
    except ValueError:
        raise error.ResponseError(
            _(b'unexpected response from remote server:'), l
        )
    if resp == 1:
        raise error.Abort(_(b'operation forbidden by server'))
    elif resp == 2:
        raise error.Abort(_(b'locking the remote repository failed'))
    elif resp != 0:
        raise error.Abort(_(b'the server sent an unknown error code'))

    l = fp.readline()
    try:
        filecount, bytecount = map(int, l.split(b' ', 1))
    except (ValueError, TypeError):
        raise error.ResponseError(
            _(b'unexpected response from remote server:'), l
        )

    with repo.lock():
        consumev1(repo, fp, filecount, bytecount)
        repo.requirements = new_stream_clone_requirements(
            repo.requirements,
            requirements,
        )
        repo.svfs.options = localrepo.resolvestorevfsoptions(
            repo.ui, repo.requirements, repo.features
        )
        scmutil.writereporequirements(repo)
        nodemap.post_stream_cleanup(repo)

        if rbranchmap:
            repo._branchcaches.replace(repo, rbranchmap)

        repo.invalidate()


def allowservergeneration(repo) -> bool:
    """Whether streaming clones are allowed from the server."""
    if repository.REPO_FEATURE_STREAM_CLONE not in repo.features:
        return False

    if not repo.ui.configbool(b'server', b'uncompressed', untrusted=True):
        return False

    # The way stream clone works makes it impossible to hide secret changesets.
    # So don't allow this by default.
    secret = phases.hassecret(repo)
    if secret:
        return repo.ui.configbool(b'server', b'uncompressedallowsecret')

    return True


# This is it's own function so extensions can override it.
def _walkstreamfiles(
    repo, matcher=None, phase: bool = False, obsolescence: bool = False
):
    return repo.store.walk(matcher, phase=phase, obsolescence=obsolescence)


def _report_transferred(
    repo, start_time: float, file_count: int, byte_count: int
):
    """common utility to report time it took to apply the stream bundle"""
    elapsed = util.timer() - start_time
    if elapsed <= 0:
        elapsed = 0.001
    m = _(b'stream-cloned %d files / %s in %.1f seconds (%s/sec)\n')
    m %= (
        file_count,
        util.bytecount(byte_count),
        elapsed,
        util.bytecount(byte_count / elapsed),
    )
    repo.ui.status(m)


def generatev1(repo) -> tuple[int, int, Iterator[bytes]]:
    """Emit content for version 1 of a streaming clone.

    This returns a 3-tuple of (file count, byte size, data iterator).

    The data iterator consists of N entries for each file being transferred.
    Each file entry starts as a line with the file name and integer size
    delimited by a null byte.

    The raw file data follows. Following the raw file data is the next file
    entry, or EOF.

    When used on the wire protocol, an additional line indicating protocol
    success will be prepended to the stream. This function is not responsible
    for adding it.

    This function will obtain a repository lock to ensure a consistent view of
    the store is captured. It therefore may raise LockError.
    """
    entries = []
    total_bytes = 0
    # Get consistent snapshot of repo, lock during scan.
    with repo.lock():
        repo.ui.debug(b'scanning\n')
        _test_sync_point_walk_1_2(repo)
        for entry in _walkstreamfiles(repo):
            for f in entry.files():
                file_size = f.file_size(repo.store.vfs)
                if file_size:
                    entries.append((f.unencoded_path, file_size))
                    total_bytes += file_size
        _test_sync_point_walk_3(repo)
    _test_sync_point_walk_4(repo)

    repo.ui.debug(
        b'%d files, %d bytes to transfer\n' % (len(entries), total_bytes)
    )

    svfs = repo.svfs
    debugflag = repo.ui.debugflag

    def emitrevlogdata() -> Iterator[bytes]:
        for name, size in entries:
            if debugflag:
                repo.ui.debug(b'sending %s (%d bytes)\n' % (name, size))
            # partially encode name over the wire for backwards compat
            yield b'%s\0%d\n' % (store.encodedir(name), size)
            # auditing at this stage is both pointless (paths are already
            # trusted by the local repo) and expensive
            with svfs(name, b'rb', auditpath=False) as fp:
                if size <= 65536:
                    yield fp.read(size)
                else:
                    yield from util.filechunkiter(fp, limit=size)

    return len(entries), total_bytes, emitrevlogdata()


def generatev1wireproto(repo) -> Iterator[bytes]:
    """Emit content for version 1 of streaming clone suitable for the wire.

    This is the data output from ``generatev1()`` with 2 header lines. The
    first line indicates overall success. The 2nd contains the file count and
    byte size of payload.

    The success line contains "0" for success, "1" for stream generation not
    allowed, and "2" for error locking the repository (possibly indicating
    a permissions error for the server process).
    """
    if not allowservergeneration(repo):
        yield b'1\n'
        return

    try:
        filecount, bytecount, it = generatev1(repo)
    except error.LockError:
        yield b'2\n'
        return

    # Indicates successful response.
    yield b'0\n'
    yield b'%d %d\n' % (filecount, bytecount)
    yield from it


def generatebundlev1(
    repo, compression: bytes = b'UN'
) -> tuple[set[bytes], Iterator[bytes]]:
    """Emit content for version 1 of a stream clone bundle.

    The first 4 bytes of the output ("HGS1") denote this as stream clone
    bundle version 1.

    The next 2 bytes indicate the compression type. Only "UN" is currently
    supported.

    The next 16 bytes are two 64-bit big endian unsigned integers indicating
    file count and byte count, respectively.

    The next 2 bytes is a 16-bit big endian unsigned short declaring the length
    of the requirements string, including a trailing \0. The following N bytes
    are the requirements string, which is ASCII containing a comma-delimited
    list of repo requirements that are needed to support the data.

    The remaining content is the output of ``generatev1()`` (which may be
    compressed in the future).

    Returns a tuple of (requirements, data generator).
    """
    if compression != b'UN':
        raise ValueError('we do not support the compression argument yet')

    requirements = streamed_requirements(repo)
    requires = b','.join(sorted(requirements))

    def gen() -> Iterator[bytes]:
        yield b'HGS1'
        yield compression

        filecount, bytecount, it = generatev1(repo)
        repo.ui.status(
            _(b'writing %d bytes for %d files\n') % (bytecount, filecount)
        )

        yield struct.pack(b'>QQ', filecount, bytecount)
        yield struct.pack(b'>H', len(requires) + 1)
        yield requires + b'\0'

        # This is where we'll add compression in the future.
        assert compression == b'UN'

        progress = repo.ui.makeprogress(
            _(b'bundle'), total=bytecount, unit=_(b'bytes')
        )
        progress.update(0)

        for chunk in it:
            progress.increment(step=len(chunk))
            yield chunk

        progress.complete()

    return requirements, gen()


def consumev1(repo, fp, filecount: int, bytecount: int) -> None:
    """Apply the contents from version 1 of a streaming clone file handle.

    This takes the output from "stream_out" and applies it to the specified
    repository.

    Like "stream_out," the status line added by the wire protocol is not
    handled by this function.
    """
    with repo.lock():
        repo.ui.status(
            _(b'%d files to transfer, %s of data\n')
            % (filecount, util.bytecount(bytecount))
        )
        progress = repo.ui.makeprogress(
            _(b'clone'), total=bytecount, unit=_(b'bytes')
        )
        progress.update(0)
        start = util.timer()

        # TODO: get rid of (potential) inconsistency
        #
        # If transaction is started and any @filecache property is
        # changed at this point, it causes inconsistency between
        # in-memory cached property and streamclone-ed file on the
        # disk. Nested transaction prevents transaction scope "clone"
        # below from writing in-memory changes out at the end of it,
        # even though in-memory changes are discarded at the end of it
        # regardless of transaction nesting.
        #
        # But transaction nesting can't be simply prohibited, because
        # nesting occurs also in ordinary case (e.g. enabling
        # clonebundles).

        total_file_count = 0
        with repo.transaction(b'clone'):
            with repo.svfs.backgroundclosing(repo.ui, expectedcount=filecount):
                for i in range(filecount):
                    # XXX doesn't support '\n' or '\r' in filenames
                    if hasattr(fp, 'readline'):
                        l = fp.readline()
                    else:
                        # inline clonebundles use a chunkbuffer, so no readline
                        # --> this should be small anyway, the first line
                        # only contains the size of the bundle
                        l_buf = []
                        while not (l_buf and l_buf[-1] == b'\n'):
                            l_buf.append(fp.read(1))
                        l = b''.join(l_buf)
                    try:
                        name, size = l.split(b'\0', 1)
                        size = int(size)
                    except (ValueError, TypeError):
                        raise error.ResponseError(
                            _(b'unexpected response from remote server:'), l
                        )
                    if repo.ui.debugflag:
                        repo.ui.debug(
                            b'adding %s (%s)\n' % (name, util.bytecount(size))
                        )
                    # for backwards compat, name was partially encoded
                    path = store.decodedir(name)
                    with repo.svfs(path, b'w', backgroundclose=True) as ofp:
                        total_file_count += 1
                        for chunk in util.filechunkiter(fp, limit=size):
                            progress.increment(step=len(chunk))
                            ofp.write(chunk)

            # force @filecache properties to be reloaded from
            # streamclone-ed file at next access
            repo.invalidate(clearfilecache=True)

        progress.complete()
        _report_transferred(repo, start, total_file_count, bytecount)


def readbundle1header(fp) -> tuple[int, int, set[bytes]]:
    compression = fp.read(2)
    if compression != b'UN':
        raise error.Abort(
            _(
                b'only uncompressed stream clone bundles are '
                b'supported; got %s'
            )
            % compression
        )

    filecount, bytecount = struct.unpack(b'>QQ', fp.read(16))
    requireslen = struct.unpack(b'>H', fp.read(2))[0]
    requires = fp.read(requireslen)

    if not requires.endswith(b'\0'):
        raise error.Abort(
            _(
                b'malformed stream clone bundle: '
                b'requirements not properly encoded'
            )
        )

    requirements = set(requires.rstrip(b'\0').split(b','))

    return filecount, bytecount, requirements


def applybundlev1(repo, fp) -> None:
    """Apply the content from a stream clone bundle version 1.

    We assume the 4 byte header has been read and validated and the file handle
    is at the 2 byte compression identifier.
    """
    if len(repo):
        raise error.Abort(
            _(b'cannot apply stream clone bundle on non-empty repo')
        )

    filecount, bytecount, requirements = readbundle1header(fp)
    missingreqs = requirements - repo.supported
    if missingreqs:
        raise error.Abort(
            _(b'unable to apply stream clone: unsupported format: %s')
            % b', '.join(sorted(missingreqs))
        )

    consumev1(repo, fp, filecount, bytecount)
    nodemap.post_stream_cleanup(repo)


class streamcloneapplier:
    """Class to manage applying streaming clone bundles.

    We need to wrap ``applybundlev1()`` in a dedicated type to enable bundle
    readers to perform bundle type-specific functionality.
    """

    def __init__(self, fh) -> None:
        self._fh = fh

    def apply(self, repo) -> None:
        return applybundlev1(repo, self._fh)


# type of file to stream
_fileappend = 0  # append only file
_filefull = 1  # full snapshot file

# Source of the file
_srcstore = b's'  # store (svfs)
_srccache = b'c'  # cache (cache)


# This is it's own function so extensions can override it.
def _walkstreamfullstorefiles(repo) -> list[bytes]:
    """list snapshot file from the store"""
    fnames = []
    if not repo.publishing():
        fnames.append(b'phaseroots')
    return fnames


def _filterfull(entry, copy, vfsmap):
    """actually copy the snapshot files"""
    src, name, ftype, data = entry
    if ftype != _filefull:
        return entry
    return (src, name, ftype, copy(vfsmap[src].join(name)))


class VolatileManager:
    """Manage temporary backups of volatile files during stream clone.

    This class will keep open file handles for the volatile files, writing the
    smaller ones on disk if the number of open file handles grow too much.

    This should be used as a Python context, the file handles and copies will
    be discarded when exiting the context.

    The preservation can be done by calling the object on the real path
    (encoded full path).

    Valid filehandles for any file should be retrieved by calling `open(path)`.
    """

    # arbitrarily picked as "it seemed fine" and much higher than the current
    # usage.  The Windows value of 2 is actually 1 file open at a time, due to
    # the `flush_count = self.MAX_OPEN // 2` and `self.MAX_OPEN - 1` threshold
    # for flushing to disk in __call__().
    MAX_OPEN = 2 if pycompat.iswindows else 100

    def __init__(self) -> None:
        self._counter = 0
        self._volatile_fps = None
        self._copies = None
        self._dst_dir = None

    def __enter__(self):
        if self._counter == 0:
            assert self._volatile_fps is None
            self._volatile_fps = {}
        self._counter += 1
        return self

    def __exit__(self, *args, **kwars):
        """discard all backups"""
        self._counter -= 1
        if self._counter == 0:
            for _size, fp in self._volatile_fps.values():
                fp.close()
            self._volatile_fps = None
            if self._copies is not None:
                for tmp in self._copies.values():
                    util.tryunlink(tmp)
                util.tryrmdir(self._dst_dir)
            self._copies = None
            self._dst_dir = None
            assert self._volatile_fps is None
            assert self._copies is None
            assert self._dst_dir is None

    def _init_tmp_copies(self) -> None:
        """prepare a temporary directory to save volatile files

        This will be used as backup if we have too many files open"""
        assert 0 < self._counter
        assert self._copies is None
        assert self._dst_dir is None
        self._copies = {}
        self._dst_dir = pycompat.mkdtemp(prefix=b'hg-clone-')

    def _flush_some_on_disk(self) -> None:
        """move some of the open files to tempory files on disk"""
        if self._copies is None:
            self._init_tmp_copies()
        flush_count = self.MAX_OPEN // 2
        for src, (size, fp) in sorted(self._volatile_fps.items())[:flush_count]:
            prefix = os.path.basename(src)
            fd, dst = pycompat.mkstemp(prefix=prefix, dir=self._dst_dir)
            self._copies[src] = dst
            os.close(fd)
            # we no longer hardlink, but on the other hand we rarely do this,
            # and we do it for the smallest file only and not at all in the
            # common case.
            with open(dst, 'wb') as bck:
                fp.seek(0)
                bck.write(fp.read())
            del self._volatile_fps[src]
            fp.close()

    def _keep_one(self, src: bytes) -> int:
        """preserve an open file handle for a given path"""
        # store the file quickly to ensure we close it if any error happens
        _, fp = self._volatile_fps[src] = (None, open(src, 'rb'))
        fp.seek(0, os.SEEK_END)
        size = fp.tell()
        self._volatile_fps[src] = (size, fp)
        return size

    def __call__(self, src: bytes) -> None:
        """preserve the volatile file at src"""
        assert 0 < self._counter
        if len(self._volatile_fps) >= (self.MAX_OPEN - 1):
            self._flush_some_on_disk()
        self._keep_one(src)

    def try_keep(self, src: bytes) -> int | None:
        """record a volatile file and returns it size

        return None if the file does not exists.

        Used for cache file that are not lock protected.
        """
        assert 0 < self._counter
        if len(self._volatile_fps) >= (self.MAX_OPEN - 1):
            self._flush_some_on_disk()
        try:
            return self._keep_one(src)
        except OSError as err:
            if err.errno not in (errno.ENOENT, errno.EPERM):
                raise
            return None

    @contextlib.contextmanager
    def open(self, src):
        assert 0 < self._counter
        entry = self._volatile_fps.get(src)
        if entry is not None:
            _size, fp = entry
            fp.seek(0)
            yield fp
        else:
            if self._copies is None:
                actual_path = src
            else:
                actual_path = self._copies.get(src, src)
            with open(actual_path, 'rb') as fp:
                yield fp


def _makemap(repo):
    """make a (src -> vfs) map for the repo"""
    vfsmap = {
        _srcstore: repo.svfs,
        _srccache: repo.cachevfs,
    }
    # we keep repo.vfs out of the on purpose, ther are too many danger there
    # (eg: .hg/hgrc)
    assert repo.vfs not in vfsmap.values()

    return vfsmap


def _emit2(repo, entries):
    """actually emit the stream bundle"""
    vfsmap = _makemap(repo)
    # we keep repo.vfs out of the on purpose, ther are too many danger there
    # (eg: .hg/hgrc),
    #
    # this assert is duplicated (from _makemap) as author might think this is
    # fine, while this is really not fine.
    if repo.vfs in vfsmap.values():
        raise error.ProgrammingError(
            'repo.vfs must not be added to vfsmap for security reasons'
        )

    # translate the vfs one
    entries = [(vfs_key, vfsmap[vfs_key], e) for (vfs_key, e) in entries]
    _test_sync_point_walk_1_2(repo)

    max_linkrev = len(repo)
    file_count = totalfilesize = 0
    with VolatileManager() as volatiles:
        # make sure we preserve volatile files
        with util.nogc():
            # record the expected size of every file
            for k, vfs, e in entries:
                e.preserve_volatiles(vfs, volatiles)
                for f in e.files():
                    file_count += 1
                    totalfilesize += f.file_size(vfs)

        progress = repo.ui.makeprogress(
            _(b'bundle'), total=totalfilesize, unit=_(b'bytes')
        )
        progress.update(0)
        with progress:
            # the first yield release the lock on the repository
            yield file_count, totalfilesize
            totalbytecount = 0

            for src, vfs, e in entries:
                entry_streams = e.get_streams(
                    repo=repo,
                    vfs=vfs,
                    volatiles=volatiles,
                    max_changeset=max_linkrev,
                    preserve_file_count=True,
                )
                for name, stream, size in entry_streams:
                    yield src
                    yield util.uvarintencode(len(name))
                    yield util.uvarintencode(size)
                    yield name
                    bytecount = 0
                    for chunk in stream:
                        bytecount += len(chunk)
                        totalbytecount += len(chunk)
                        progress.update(totalbytecount)
                        yield chunk
                    if bytecount != size:
                        # Would most likely be caused by a race due to `hg
                        # strip` or a revlog split
                        msg = _(
                            b'clone could only read %d bytes from %s, but '
                            b'expected %d bytes'
                        )
                        raise error.Abort(msg % (bytecount, name, size))


def _emit3(repo, entries) -> Iterator[bytes | None]:
    """actually emit the stream bundle (v3)"""
    vfsmap = _makemap(repo)
    # we keep repo.vfs out of the map on purpose, ther are too many dangers
    # there (eg: .hg/hgrc),
    #
    # this assert is duplicated (from _makemap) as authors might think this is
    # fine, while this is really not fine.
    if repo.vfs in vfsmap.values():
        raise error.ProgrammingError(
            'repo.vfs must not be added to vfsmap for security reasons'
        )

    # translate the vfs once
    # we only turn this into a list for the `_test_sync`, this is not ideal
    base_entries = list(entries)
    _test_sync_point_walk_1_2(repo)
    entries = []
    with VolatileManager() as volatiles:
        # make sure we preserve volatile files
        for vfs_key, e in base_entries:
            vfs = vfsmap[vfs_key]
            any_files = True
            if e.maybe_volatile:
                any_files = False
                e.preserve_volatiles(vfs, volatiles)
                for f in e.files():
                    if f.is_volatile:
                        # record the expected size under lock
                        f.file_size(vfs)
                    any_files = True
            if any_files:
                entries.append((vfs_key, vfsmap[vfs_key], e))

        total_entry_count = len(entries)

        max_linkrev = len(repo)
        progress = repo.ui.makeprogress(
            _(b'bundle'),
            total=total_entry_count,
            unit=_(b'entry'),
        )
        progress.update(0)
        # the first yield release the lock on the repository
        yield None
        with progress:
            yield util.uvarintencode(total_entry_count)

            for src, vfs, e in entries:
                entry_streams = e.get_streams(
                    repo=repo,
                    vfs=vfs,
                    volatiles=volatiles,
                    max_changeset=max_linkrev,
                )
                yield util.uvarintencode(len(entry_streams))
                for name, stream, size in entry_streams:
                    yield src
                    yield util.uvarintencode(len(name))
                    yield util.uvarintencode(size)
                    yield name
                    yield from stream
                progress.increment()


def _test_sync_point_walk_1_2(repo):
    """a function for synchronisation during tests

    Triggered after gather entry, but before starting to process/preserve them
    under lock.

    (on v1 is triggered before the actual walk start)
    """


def _test_sync_point_walk_3(repo):
    """a function for synchronisation during tests

    Triggered right before releasing the lock, but after computing what need
    needed to compute under lock.
    """


def _test_sync_point_walk_4(repo):
    """a function for synchronisation during tests

    Triggered right after releasing the lock.
    """


# not really a StoreEntry, but close enough
class CacheEntry(store.SimpleStoreEntry):
    """Represent an entry for Cache files

    It has special logic to preserve cache file early and accept optional
    presence.


    (Yes... this is not really a StoreEntry, but close enough. We could have a
    BaseEntry base class, bbut the store one would be identical)
    """

    def __init__(self, entry_path) -> None:
        super().__init__(
            entry_path,
            # we will directly deal with that in `setup_cache_file`
            is_volatile=True,
        )

    def preserve_volatiles(self, vfs, volatiles) -> None:
        self._file_size = volatiles.try_keep(vfs.join(self._entry_path))
        if self._file_size is None:
            self._files = []
        else:
            assert self._is_volatile
            self._files = [
                CacheFile(
                    unencoded_path=self._entry_path,
                    file_size=self._file_size,
                    is_volatile=self._is_volatile,
                )
            ]

    def files(self) -> list[store.StoreFile]:
        if self._files is None:
            self._files = [
                CacheFile(
                    unencoded_path=self._entry_path,
                    is_volatile=self._is_volatile,
                )
            ]
        return super().files()


class CacheFile(store.StoreFile):
    # inform the "copy/hardlink" version that this file might be missing
    # without consequences.
    optional: bool = True


def _entries_walk(repo, includes, excludes, includeobsmarkers: bool):
    """emit a seris of files information useful to clone a repo

    return (vfs-key, entry) iterator

    Where `entry` is StoreEntry. (used even for cache entries)
    """
    assert repo._currentlock(repo._lockref) is not None

    matcher = None
    if includes or excludes:
        matcher = narrowspec.match(repo.root, includes, excludes)

    phase = not repo.publishing()
    # Python is getting crazy at all the small container we creates, disabling
    # the gc while we do so helps performance a lot.
    with util.nogc():
        entries = _walkstreamfiles(
            repo,
            matcher,
            phase=phase,
            obsolescence=includeobsmarkers,
        )
        for entry in entries:
            yield (_srcstore, entry)

        for name in cacheutil.cachetocopy(repo):
            if repo.cachevfs.exists(name):
                # not really a StoreEntry, but close enough
                yield (_srccache, CacheEntry(entry_path=name))


def generatev2(repo, includes, excludes, includeobsmarkers: bool):
    """Emit content for version 2 of a streaming clone.

    the data stream consists the following entries:
    1) A char representing the file destination (eg: store or cache)
    2) A varint containing the length of the filename
    3) A varint containing the length of file data
    4) N bytes containing the filename (the internal, store-agnostic form)
    5) N bytes containing the file data

    Returns a 3-tuple of (file count, file size, data iterator).
    """

    with repo.lock():
        repo.ui.debug(b'scanning\n')

        entries = _entries_walk(
            repo,
            includes=includes,
            excludes=excludes,
            includeobsmarkers=includeobsmarkers,
        )
        chunks = _emit2(repo, entries)
        first = next(chunks)
        file_count, total_file_size = first
        _test_sync_point_walk_3(repo)
    _test_sync_point_walk_4(repo)

    return file_count, total_file_size, chunks


def generatev3(
    repo, includes, excludes, includeobsmarkers: bool
) -> Iterator[bytes | None]:
    """Emit content for version 3 of a streaming clone.

    the data stream consists the following:
    1) A varint E containing the number of entries (can be 0), then E entries follow
    2) For each entry:
    2.1) The number of files in this entry (can be 0, but typically 1 or 2)
    2.2) For each file:
    2.2.1) A char representing the file destination (eg: store or cache)
    2.2.2) A varint N containing the length of the filename
    2.2.3) A varint M containing the length of file data
    2.2.4) N bytes containing the filename (the internal, store-agnostic form)
    2.2.5) M bytes containing the file data

    Returns the data iterator.

    XXX This format is experimental and subject to change. Here is a
    XXX non-exhaustive list of things this format could do or change:

    - making it easier to write files in parallel
    - holding the lock for a shorter time
    - improving progress information
    - ways to adjust the number of expected entries/files ?
    """

    # Python is getting crazy at all the small container we creates while
    # considering the files to preserve, disabling the gc while we do so helps
    # performance a lot.
    with repo.lock(), util.nogc():
        repo.ui.debug(b'scanning\n')

        entries = _entries_walk(
            repo,
            includes=includes,
            excludes=excludes,
            includeobsmarkers=includeobsmarkers,
        )
        chunks = _emit3(repo, list(entries))
        first = next(chunks)
        assert first is None
        _test_sync_point_walk_3(repo)
    _test_sync_point_walk_4(repo)

    return chunks


@contextlib.contextmanager
def nested(*ctxs):
    this = ctxs[0]
    rest = ctxs[1:]
    with this:
        if rest:
            with nested(*rest):
                yield
        else:
            yield


class V2Report:
    """a small class to track the data we saw within the stream"""

    def __init__(self):
        self.byte_count = 0


def consumev2(repo, fp, filecount: int, filesize: int) -> None:
    """Apply the contents from a version 2 streaming clone.

    Data is read from an object that only needs to provide a ``read(size)``
    method.
    """
    with repo.lock():
        repo.ui.status(
            _(b'%d files to transfer, %s of data\n')
            % (filecount, util.bytecount(filesize))
        )
        progress = repo.ui.makeprogress(
            _(b'clone'),
            total=filesize,
            unit=_(b'bytes'),
        )
        start = util.timer()
        report = V2Report()

        vfsmap = _makemap(repo)
        # we keep repo.vfs out of the on purpose, ther are too many danger
        # there (eg: .hg/hgrc),
        #
        # this assert is duplicated (from _makemap) as author might think this
        # is fine, while this is really not fine.
        if repo.vfs in vfsmap.values():
            raise error.ProgrammingError(
                'repo.vfs must not be added to vfsmap for security reasons'
            )

        cpu_profile = scmutil.get_resource_profile(repo.ui, b'cpu')
        mem_profile = scmutil.get_resource_profile(repo.ui, b'memory')
        threaded = repo.ui.configbool(
            b"worker", b"parallel-stream-bundle-processing"
        )
        num_writer = repo.ui.configint(
            b"worker",
            b"parallel-stream-bundle-processing.num-writer",
        )
        if num_writer <= 0:
            num_writer = DEFAULT_NUM_WRITER[cpu_profile]
        memory_target = repo.ui.configbytes(
            b"worker",
            b"parallel-stream-bundle-processing.memory-target",
        )
        if memory_target < 0:
            memory_target = None
        elif memory_target == 0:
            memory_target = DEFAULT_MEMORY_TARGET[mem_profile]
        with repo.transaction(b'clone'):
            ctxs = (vfs.backgroundclosing(repo.ui) for vfs in vfsmap.values())
            with nested(*ctxs):
                workers = []
                info_queue = None
                data_queue = None
                mark_used = None
                try:
                    if not threaded:
                        fc = _FileChunker
                        raw_data = fp
                    else:
                        fc = _ThreadSafeFileChunker
                        data_queue = _DataQueue(memory_target=memory_target)
                        if memory_target is not None:
                            mark_used = data_queue.mark_used
                        raw_data = util.chunkbuffer(data_queue)

                        w = threading.Thread(
                            target=data_queue.fill_from,
                            args=(fp,),
                        )
                        workers.append(w)
                        w.start()
                    files = _v2_parse_files(
                        repo,
                        raw_data,
                        vfsmap,
                        filecount,
                        progress,
                        report,
                        file_chunker=fc,
                        mark_used=mark_used,
                    )
                    if not threaded:
                        _write_files(files)
                    else:
                        info_queue = _FileInfoQueue(files)

                        for __ in range(num_writer):
                            w = threading.Thread(
                                target=_write_files,
                                args=(info_queue,),
                            )
                            workers.append(w)
                            w.start()
                        info_queue.fill()
                except:  # re-raises
                    if data_queue is not None:
                        data_queue.abort()
                    raise
                finally:
                    # shut down all the workers
                    if info_queue is not None:
                        # this is strictly speaking one too many worker for
                        # this queu, but closing too many is not a problem.
                        info_queue.close(len(workers))
                    for w in workers:
                        w.join()

            # force @filecache properties to be reloaded from
            # streamclone-ed file at next access
            repo.invalidate(clearfilecache=True)

        progress.complete()
        # acknowledge the end of the bundle2 part, this help aligning
        # sequential and parallel behavior.
        remains = fp.read(1)
        assert not remains
        _report_transferred(repo, start, filecount, report.byte_count)


# iterator of chunk of bytes that constitute a file content.
FileChunksT = Iterator[bytes]
# Contains the information necessary to write stream file on disk
FileInfoT = tuple[
    bytes,  # real fs path
    Optional[int],  # permission to give to chmod
    FileChunksT,  # content
]


class _Queue:
    """a reimplementation of queue.Queue which doesn't use thread.Condition"""

    def __init__(self):
        self._queue = collections.deque()

        # the "_lock" protect manipulation of the "_queue" deque
        # the "_wait" is used to have the "get" thread waits for the
        # "put" thread when the queue is empty.
        #
        # This is similar to the "threading.Condition", but without the absurd
        # slowness of the stdlib implementation.
        #
        # the "_wait" is always released while holding the "_lock".
        self._lock = threading.Lock()
        self._wait = threading.Lock()

    def put(self, item):
        with self._lock:
            self._queue.append(item)
            # if anyone is waiting on item, unblock it.
            if self._wait.locked():
                self._wait.release()

    def get(self):
        with self._lock:
            while len(self._queue) == 0:
                # "arm"  the waiting lock
                self._wait.acquire(blocking=False)
                # release the lock to let other touch the queue
                # (especially the put call we wait on)
                self._lock.release()
                # wait for for a `put` call to release the lock
                self._wait.acquire()
                # grab the lock to look at a possible available value
                self._lock.acquire()
                # disarm the lock if necessary.
                #
                # If the queue only constains one item, keep the _wait lock
                # armed, as there is no need to wake another waiter anyway.
                if self._wait.locked() and len(self._queue) > 1:
                    self._wait.release()
            return self._queue.popleft()


class _DataQueue:
    """A queue passing data from the bundle stream to other thread

    It has a "memory_target" optional parameter to avoid buffering too much
    information. The implementation is not exact and the memory target might be
    exceed for a time in some situation.
    """

    def __init__(self, memory_target=None):
        self._q = _Queue()
        self._abort = False
        self._memory_target = memory_target
        if self._memory_target is not None and self._memory_target <= 0:
            raise error.ProgrammingError("memory target should be > 0")

        # the "_lock" protect manipulation of the _current_used" variable
        # the "_wait" is used to have the "reading" thread waits for the
        # "using" thread when the buffer is full.
        #
        # This is similar to the "threading.Condition", but without the absurd
        # slowness of the stdlib implementation.
        #
        # the "_wait" is always released while holding the "_lock".
        self._lock = threading.Lock()
        self._wait = threading.Lock()
        # only the stream reader touch this, it is find to touch without the lock
        self._current_read = 0
        # do not touch this without the lock
        self._current_used = 0

    def _has_free_space(self):
        """True if more data can be read without further exceeding memory target

        Must be called under the lock.
        """
        if self._memory_target is None:
            # Ideally we should not even get into the locking business in that
            # case, but we keep the implementation simple for now.
            return True
        return (self._current_read - self._current_used) < self._memory_target

    def mark_used(self, offset):
        """Notify we have used the buffer up to "offset"

        This is meant to be used from another thread than the one filler the queue.
        """
        if self._memory_target is not None:
            with self._lock:
                if offset > self._current_used:
                    self._current_used = offset
                # If the reader is waiting for room, unblock it.
                if self._wait.locked() and self._has_free_space():
                    self._wait.release()

    def fill_from(self, data):
        """fill the data queue from a bundle2 part object

        This is meant to be called by the data reading thread
        """
        q = self._q
        try:
            for item in data:
                self._current_read += len(item)
                q.put(item)
                if self._abort:
                    break
                if self._memory_target is not None:
                    with self._lock:
                        while not self._has_free_space():
                            # make sure the _wait lock is locked
                            # this is done under lock, so there case be no race with the release logic
                            self._wait.acquire(blocking=False)
                            self._lock.release()
                            # acquiring the lock will block until some other thread release it.
                            self._wait.acquire()
                            # lets dive into the locked section again
                            self._lock.acquire()
                            # make sure we release the lock we just grabed if
                            # needed.
                            if self._wait.locked():
                                self._wait.release()
        finally:
            q.put(None)

    def __iter__(self):
        """Iterate over the bundle chunkgs

        This is meant to be called by the data parsing thread."""
        q = self._q
        while (i := q.get()) is not None:
            yield i
            if self._abort:
                break

    def abort(self):
        """stop the data-reading thread and interrupt the comsuming iteration

        This is meant to be called on errors.
        """
        self._abort = True
        self._q.put(None)
        if self._memory_target is not None:
            with self._lock:
                # make sure we unstuck the reader thread.
                if self._wait.locked():
                    self._wait.release()


class _FileInfoQueue:
    """A thread-safe queue to passer parsed file information to the writers"""

    def __init__(self, info: Iterable[FileInfoT]):
        self._info = info
        self._q = _Queue()

    def fill(self):
        """iterate over the parsed information to file the queue

        This is meant to be call from the thread parsing the stream information.
        """
        q = self._q
        for i in self._info:
            q.put(i)

    def close(self, number_worker):
        """signal all the workers that we no longer have any file info coming

        Called from the thread parsing the stream information (and/or the main
        thread if different).
        """
        for __ in range(number_worker):
            self._q.put(None)

    def __iter__(self):
        """iterate over the available file info

        This is meant to be called from the writer threads.
        """
        q = self._q
        while (i := q.get()) is not None:
            yield i


class _FileChunker:
    """yield the chunk that constitute a file

    This class exists as the counterpart of the threaded version and
    would not be very useful on its own.
    """

    def __init__(
        self,
        fp: bundle2mod.unbundlepart,
        data_len: int,
        progress: scmutil.progress,
        report: V2Report,
        mark_used: Callable[[int], None] | None = None,
    ):
        self.report = report
        self.progress = progress
        self._chunks = util.filechunkiter(fp, limit=data_len)

    def fill(self) -> None:
        """Do nothing in non-threading context"""

    def __iter__(self) -> FileChunksT:
        for chunk in self._chunks:
            self.report.byte_count += len(chunk)
            self.progress.increment(step=len(chunk))
            yield chunk


class _ThreadSafeFileChunker(_FileChunker):
    """yield the chunk that constitute a file

    Make sure you  call the "fill" function in the main thread to read the
    right data at the right time.
    """

    def __init__(
        self,
        fp: bundle2mod.unbundlepart,
        data_len: int,
        progress: scmutil.progress,
        report: V2Report,
        mark_used: Callable[[int], None] | None = None,
    ):
        super().__init__(fp, data_len, progress, report)
        self._fp = fp
        self._queue = _Queue()
        self._mark_used = mark_used

    def fill(self) -> None:
        """fill the file chunker queue with data read from the stream

        This is meant to be called from the thread parsing information (and
        consuming the stream data).
        """
        try:
            for chunk in super().__iter__():
                offset = self._fp.tell()
                self._queue.put((chunk, offset))
        finally:
            self._queue.put(None)

    def __iter__(self) -> FileChunksT:
        """Iterate over all the file chunk

        This is meant to be called from the writer threads.
        """
        while (info := self._queue.get()) is not None:
            chunk, offset = info
            if self._mark_used is not None:
                self._mark_used(offset)
            yield chunk


def _trivial_file(
    chunk: bytes,
    mark_used: Callable[[int], None] | None,
    offset: int,
) -> FileChunksT:
    """used for single chunk file,"""
    if mark_used is not None:
        mark_used(offset)
    yield chunk


def _v2_parse_files(
    repo,
    fp: bundle2mod.unbundlepart,
    vfs_map,
    file_count: int,
    progress: scmutil.progress,
    report: V2Report,
    file_chunker: type[_FileChunker] = _FileChunker,
    mark_used: Callable[[int], None] | None = None,
) -> Iterator[FileInfoT]:
    """do the "stream-parsing" part of stream v2

    The parsed information are yield result for consumption by the "writer"
    """
    known_dirs = set()  # set of directory that we know to exists
    progress.update(0)
    for i in range(file_count):
        src = util.readexactly(fp, 1)
        namelen = util.uvarintdecodestream(fp)
        datalen = util.uvarintdecodestream(fp)

        name = util.readexactly(fp, namelen)

        if repo.ui.debugflag:
            repo.ui.debug(
                b'adding [%s] %s (%s)\n' % (src, name, util.bytecount(datalen))
            )
        vfs = vfs_map[src]
        path, mode = vfs.prepare_streamed_file(name, known_dirs)
        if datalen <= util.DEFAULT_FILE_CHUNK:
            c = fp.read(datalen)
            offset = fp.tell()
            report.byte_count += len(c)
            progress.increment(step=len(c))
            chunks = _trivial_file(c, mark_used, offset)
            yield (path, mode, iter(chunks))
        else:
            chunks = file_chunker(
                fp,
                datalen,
                progress,
                report,
                mark_used=mark_used,
            )
            yield (path, mode, iter(chunks))
            # make sure we read all the chunk before moving to the next file
            chunks.fill()


def _write_files(info: Iterable[FileInfoT]):
    """write files from parsed data"""
    io_flags = os.O_WRONLY | os.O_CREAT
    if pycompat.iswindows:
        io_flags |= os.O_BINARY
    for path, mode, data in info:
        if mode is None:
            fd = os.open(path, io_flags)
        else:
            fd = os.open(path, io_flags, mode=mode)
        try:
            for chunk in data:
                written = os.write(fd, chunk)
                # write missing pieces if the write was interrupted
                while written < len(chunk):
                    written = os.write(fd, chunk[written:])
        finally:
            os.close(fd)


def consumev3(repo, fp) -> None:
    """Apply the contents from a version 3 streaming clone.

    Data is read from an object that only needs to provide a ``read(size)``
    method.
    """
    with repo.lock():
        start = util.timer()

        entrycount = util.uvarintdecodestream(fp)
        repo.ui.status(_(b'%d entries to transfer\n') % (entrycount))

        progress = repo.ui.makeprogress(
            _(b'clone'),
            total=entrycount,
            unit=_(b'entries'),
        )
        progress.update(0)
        bytes_transferred = 0

        vfsmap = _makemap(repo)
        # we keep repo.vfs out of the on purpose, there are too many dangers
        # there (eg: .hg/hgrc),
        #
        # this assert is duplicated (from _makemap) as authors might think this
        # is fine, while this is really not fine.
        if repo.vfs in vfsmap.values():
            raise error.ProgrammingError(
                'repo.vfs must not be added to vfsmap for security reasons'
            )
        total_file_count = 0
        with repo.transaction(b'clone'):
            ctxs = (vfs.backgroundclosing(repo.ui) for vfs in vfsmap.values())
            with nested(*ctxs):
                for i in range(entrycount):
                    filecount = util.uvarintdecodestream(fp)
                    if filecount == 0:
                        if repo.ui.debugflag:
                            repo.ui.debug(b'entry with no files [%d]\n' % (i))
                    total_file_count += filecount
                    for i in range(filecount):
                        src = util.readexactly(fp, 1)
                        vfs = vfsmap[src]
                        namelen = util.uvarintdecodestream(fp)
                        datalen = util.uvarintdecodestream(fp)

                        name = util.readexactly(fp, namelen)

                        if repo.ui.debugflag:
                            msg = b'adding [%s] %s (%s)\n'
                            msg %= (src, name, util.bytecount(datalen))
                            repo.ui.debug(msg)
                        bytes_transferred += datalen

                        with vfs(name, b'w') as ofp:
                            for chunk in util.filechunkiter(fp, limit=datalen):
                                ofp.write(chunk)
                    progress.increment(step=1)

            # force @filecache properties to be reloaded from
            # streamclone-ed file at next access
            repo.invalidate(clearfilecache=True)

        progress.complete()
        _report_transferred(repo, start, total_file_count, bytes_transferred)


def applybundlev2(
    repo, fp, filecount: int, filesize: int, requirements: Iterable[bytes]
) -> None:
    from . import localrepo

    missingreqs = [r for r in requirements if r not in repo.supported]
    if missingreqs:
        raise error.Abort(
            _(b'unable to apply stream clone: unsupported format: %s')
            % b', '.join(sorted(missingreqs))
        )

    with util.nogc():
        consumev2(repo, fp, filecount, filesize)

    repo.requirements = new_stream_clone_requirements(
        repo.requirements,
        requirements,
    )
    repo.svfs.options = localrepo.resolvestorevfsoptions(
        repo.ui, repo.requirements, repo.features
    )
    scmutil.writereporequirements(repo)
    nodemap.post_stream_cleanup(repo)


def applybundlev3(repo, fp, requirements: Iterable[bytes]) -> None:
    from . import localrepo

    missingreqs = [r for r in requirements if r not in repo.supported]
    if missingreqs:
        msg = _(b'unable to apply stream clone: unsupported format: %s')
        msg %= b', '.join(sorted(missingreqs))
        raise error.Abort(msg)

    consumev3(repo, fp)

    repo.requirements = new_stream_clone_requirements(
        repo.requirements,
        requirements,
    )
    repo.svfs.options = localrepo.resolvestorevfsoptions(
        repo.ui, repo.requirements, repo.features
    )
    scmutil.writereporequirements(repo)
    nodemap.post_stream_cleanup(repo)


def _copy_files(src_vfs_map, dst_vfs_map, entries, progress) -> bool:
    hardlink = [True]

    def copy_used():
        hardlink[0] = False
        progress.topic = _(b'copying')

    for k, path, optional in entries:
        src_vfs = src_vfs_map[k]
        dst_vfs = dst_vfs_map[k]
        src_path = src_vfs.join(path)
        dst_path = dst_vfs.join(path)
        # We cannot use dirname and makedirs of dst_vfs here because the store
        # encoding confuses them. See issue 6581 for details.
        dirname = os.path.dirname(dst_path)
        if not os.path.exists(dirname):
            util.makedirs(dirname)
        dst_vfs.register_file(path)
        # XXX we could use the #nb_bytes argument.
        try:
            util.copyfile(
                src_path,
                dst_path,
                hardlink=hardlink[0],
                no_hardlink_cb=copy_used,
                check_fs_hardlink=False,
            )
        except FileNotFoundError:
            if not optional:
                raise
        progress.increment()
    return hardlink[0]


def local_copy(src_repo, dest_repo) -> None:
    """copy all content from one local repository to another

    This is useful for local clone"""
    src_store_requirements = {
        r
        for r in src_repo.requirements
        if r not in requirementsmod.WORKING_DIR_REQUIREMENTS
    }
    dest_store_requirements = {
        r
        for r in dest_repo.requirements
        if r not in requirementsmod.WORKING_DIR_REQUIREMENTS
    }
    assert src_store_requirements == dest_store_requirements

    with dest_repo.lock():
        with src_repo.lock():
            # bookmark is not integrated to the streaming as it might use the
            # `repo.vfs` and they are too many sentitive data accessible
            # through `repo.vfs` to expose it to streaming clone.
            src_book_vfs = bookmarks.bookmarksvfs(src_repo)
            srcbookmarks = src_book_vfs.join(b'bookmarks')
            bm_count = 0
            if os.path.exists(srcbookmarks):
                bm_count = 1

            entries = _entries_walk(
                src_repo,
                includes=None,
                excludes=None,
                includeobsmarkers=True,
            )
            entries = list(entries)
            src_vfs_map = _makemap(src_repo)
            dest_vfs_map = _makemap(dest_repo)
            total_files = sum(len(e[1].files()) for e in entries) + bm_count
            progress = src_repo.ui.makeprogress(
                topic=_(b'linking'),
                total=total_files,
                unit=_(b'files'),
            )
            # copy  files
            #
            # We could copy the full file while the source repository is locked
            # and the other one without the lock. However, in the linking case,
            # this would also requires checks that nobody is appending any data
            # to the files while we do the clone, so this is not done yet. We
            # could do this blindly when copying files.
            files = [
                (vfs_key, f.unencoded_path, f.optional)
                for vfs_key, e in entries
                for f in e.files()
            ]
            hardlink = _copy_files(src_vfs_map, dest_vfs_map, files, progress)

            # copy bookmarks over
            if bm_count:
                dst_book_vfs = bookmarks.bookmarksvfs(dest_repo)
                dstbookmarks = dst_book_vfs.join(b'bookmarks')
                util.copyfile(srcbookmarks, dstbookmarks)
        progress.complete()
        if hardlink:
            msg = b'linked %d files\n'
        else:
            msg = b'copied %d files\n'
        src_repo.ui.debug(msg % total_files)

        with dest_repo.transaction(b"localclone") as tr:
            dest_repo.store.write(tr)

        # clean up transaction file as they do not make sense
        transaction.cleanup_undo_files(dest_repo.ui.warn, dest_repo.vfs_map)



# Mercurial built-in replacement for cvsps.
#
# Copyright 2008, Frank Kingswood <frank@kingswood-consulting.co.uk>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import functools
import os
import pickle
import re

from mercurial.i18n import _
from mercurial import (
    encoding,
    error,
    hook,
    pycompat,
    util,
)
from mercurial.utils import (
    dateutil,
    procutil,
    stringutil,
)


class logentry:
    """Class logentry has the following attributes:
    .author    - author name as CVS knows it
    .branch    - name of branch this revision is on
    .branches  - revision tuple of branches starting at this revision
    .comment   - commit message
    .commitid  - CVS commitid or None
    .date      - the commit date as a (time, tz) tuple
    .dead      - true if file revision is dead
    .file      - Name of file
    .lines     - a tuple (+lines, -lines) or None
    .parent    - Previous revision of this entry
    .rcs       - name of file as returned from CVS
    .revision  - revision number as tuple
    .tags      - list of tags on the file
    .synthetic - is this a synthetic "file ... added on ..." revision?
    .mergepoint - the branch that has been merged from (if present in
                  rlog output) or None
    .branchpoints - the branches that start at the current entry or empty
    """

    def __init__(self, **entries):
        self.synthetic = False
        self.__dict__.update(entries)

    def __repr__(self):
        items = ("%s=%r" % (k, self.__dict__[k]) for k in sorted(self.__dict__))
        return "%s(%s)" % (type(self).__name__, ", ".join(items))


class logerror(Exception):
    pass


def getrepopath(cvspath):
    """Return the repository path from a CVS path.

    >>> getrepopath(b'/foo/bar')
    '/foo/bar'
    >>> getrepopath(b'c:/foo/bar')
    '/foo/bar'
    >>> getrepopath(b':pserver:10/foo/bar')
    '/foo/bar'
    >>> getrepopath(b':pserver:10c:/foo/bar')
    '/foo/bar'
    >>> getrepopath(b':pserver:/foo/bar')
    '/foo/bar'
    >>> getrepopath(b':pserver:c:/foo/bar')
    '/foo/bar'
    >>> getrepopath(b':pserver:truc@foo.bar:/foo/bar')
    '/foo/bar'
    >>> getrepopath(b':pserver:truc@foo.bar:c:/foo/bar')
    '/foo/bar'
    >>> getrepopath(b'user@server/path/to/repository')
    '/path/to/repository'
    """
    # According to CVS manual, CVS paths are expressed like:
    # [:method:][[user][:password]@]hostname[:[port]]/path/to/repository
    #
    # CVSpath is splitted into parts and then position of the first occurrence
    # of the '/' char after the '@' is located. The solution is the rest of the
    # string after that '/' sign including it

    parts = cvspath.split(b':')
    atposition = parts[-1].find(b'@')
    start = 0

    if atposition != -1:
        start = atposition

    repopath = parts[-1][parts[-1].find(b'/', start) :]
    return repopath


def createlog(ui, directory=None, root=b"", rlog=True, cache=None):
    '''Collect the CVS rlog'''

    # Because we store many duplicate commit log messages, reusing strings
    # saves a lot of memory and pickle storage space.
    _scache = {}

    def scache(s):
        """return a shared version of a string"""
        return _scache.setdefault(s, s)

    ui.status(_(b'collecting CVS rlog\n'))

    log = []  # list of logentry objects containing the CVS state

    # patterns to match in CVS (r)log output, by state of use
    re_00 = re.compile(b'RCS file: (.+)$')
    re_01 = re.compile(b'cvs \\[r?log aborted\\]: (.+)$')
    re_02 = re.compile(b'cvs (r?log|server): (.+)\n$')
    re_03 = re.compile(
        b"(Cannot access.+CVSROOT)|(can't create temporary directory.+)$"
    )
    re_10 = re.compile(b'Working file: (.+)$')
    re_20 = re.compile(b'symbolic names:')
    re_30 = re.compile(b'\t(.+): ([\\d.]+)$')
    re_31 = re.compile(b'----------------------------$')
    re_32 = re.compile(
        b'======================================='
        b'======================================$'
    )
    re_50 = re.compile(br'revision ([\d.]+)(\s+locked by:\s+.+;)?$')
    re_60 = re.compile(
        br'date:\s+(.+);\s+author:\s+(.+);\s+state:\s+(.+?);'
        br'(\s+lines:\s+(\+\d+)?\s+(-\d+)?;)?'
        br'(\s+commitid:\s+([^;]+);)?'
        br'(.*mergepoint:\s+([^;]+);)?'
    )
    re_70 = re.compile(b'branches: (.+);$')

    file_added_re = re.compile(br'file [^/]+ was (initially )?added on branch')

    prefix = b''  # leading path to strip of what we get from CVS

    if directory is None:
        # Current working directory

        # Get the real directory in the repository
        try:
            with open(os.path.join(b'CVS', b'Repository'), 'rb') as f:
                prefix = f.read().strip()
            directory = prefix
            if prefix == b".":
                prefix = b""
        except OSError:
            raise logerror(_(b'not a CVS sandbox'))

        if prefix and not prefix.endswith(pycompat.ossep):
            prefix += pycompat.ossep

        # Use the Root file in the sandbox, if it exists
        try:
            root = util.readfile(os.path.join(b'CVS', b'Root')).strip()
        except OSError:
            pass

    if not root:
        root = encoding.environ.get(b'CVSROOT', b'')

    # read log cache if one exists
    oldlog = []
    date = None

    if cache:
        cachedir = os.path.expanduser(b'~/.hg.cvsps')
        if not os.path.exists(cachedir):
            os.mkdir(cachedir)

        # The cvsps cache pickle needs a uniquified name, based on the
        # repository location. The address may have all sort of nasties
        # in it, slashes, colons and such. So here we take just the
        # alphanumeric characters, concatenated in a way that does not
        # mix up the various components, so that
        #    :pserver:user@server:/path
        # and
        #    /pserver/user/server/path
        # are mapped to different cache file names.
        cachefile = root.split(b":") + [directory, b"cache"]
        cachefile = [b'-'.join(re.findall(br'\w+', s)) for s in cachefile if s]
        cachefile = os.path.join(
            cachedir, b'.'.join([s for s in cachefile if s])
        )

    if cache == b'update':
        try:
            ui.note(_(b'reading cvs log cache %s\n') % cachefile)
            with open(cachefile, 'rb') as fp:
                oldlog = pickle.load(fp)
                for e in oldlog:
                    if not (
                        hasattr(e, 'branchpoints')
                        and hasattr(e, 'commitid')
                        and hasattr(e, 'mergepoint')
                    ):
                        ui.status(_(b'ignoring old cache\n'))
                        oldlog = []
                        break

            ui.note(_(b'cache has %d log entries\n') % len(oldlog))
        except Exception as e:
            ui.note(_(b'error reading cache: %r\n') % e)

        if oldlog:
            date = oldlog[-1].date  # last commit date as a (time,tz) tuple
            date = dateutil.datestr(date, b'%Y/%m/%d %H:%M:%S %1%2')

    # build the CVS commandline
    cmd = [b'cvs', b'-q']
    if root:
        cmd.append(b'-d%s' % root)
        p = util.normpath(getrepopath(root))
        if not p.endswith(b'/'):
            p += b'/'
        if prefix:
            # looks like normpath replaces "" by "."
            prefix = p + util.normpath(prefix)
        else:
            prefix = p
    cmd.append([b'log', b'rlog'][rlog])
    if date:
        # no space between option and date string
        cmd.append(b'-d>%s' % date)
    cmd.append(directory)

    # state machine begins here
    tags = {}  # dictionary of revisions on current file with their tags
    branchmap = {}  # mapping between branch names and revision numbers
    rcsmap = {}
    state = 0
    store = False  # set when a new record can be appended

    cmd = [procutil.shellquote(arg) for arg in cmd]
    ui.note(_(b"running %s\n") % (b' '.join(cmd)))
    ui.debug(b"prefix=%r directory=%r root=%r\n" % (prefix, directory, root))

    pfp = procutil.popen(b' '.join(cmd), b'rb')
    peek = util.fromnativeeol(pfp.readline())
    while True:
        line = peek
        if line == b'':
            break
        peek = util.fromnativeeol(pfp.readline())
        if line.endswith(b'\n'):
            line = line[:-1]
        # ui.debug('state=%d line=%r\n' % (state, line))

        if state == 0:
            # initial state, consume input until we see 'RCS file'
            match = re_00.match(line)
            if match:
                rcs = match.group(1)
                tags = {}
                if rlog:
                    filename = util.normpath(rcs[:-2])
                    if filename.startswith(prefix):
                        filename = filename[len(prefix) :]
                    if filename.startswith(b'/'):
                        filename = filename[1:]
                    if filename.startswith(b'Attic/'):
                        filename = filename[6:]
                    else:
                        filename = filename.replace(b'/Attic/', b'/')
                    state = 2
                    continue
                state = 1
                continue
            match = re_01.match(line)
            if match:
                raise logerror(match.group(1))
            match = re_02.match(line)
            if match:
                raise logerror(match.group(2))
            if re_03.match(line):
                raise logerror(line)

        elif state == 1:
            # expect 'Working file' (only when using log instead of rlog)
            match = re_10.match(line)
            assert match, _(b'RCS file must be followed by working file')
            filename = util.normpath(match.group(1))
            state = 2

        elif state == 2:
            # expect 'symbolic names'
            if re_20.match(line):
                branchmap = {}
                state = 3

        elif state == 3:
            # read the symbolic names and store as tags
            match = re_30.match(line)
            if match:
                rev = [int(x) for x in match.group(2).split(b'.')]

                # Convert magic branch number to an odd-numbered one
                revn = len(rev)
                if revn > 3 and (revn % 2) == 0 and rev[-2] == 0:
                    rev = rev[:-2] + rev[-1:]
                rev = tuple(rev)

                if rev not in tags:
                    tags[rev] = []
                tags[rev].append(match.group(1))
                branchmap[match.group(1)] = match.group(2)

            elif re_31.match(line):
                state = 5
            elif re_32.match(line):
                state = 0

        elif state == 4:
            # expecting '------' separator before first revision
            if re_31.match(line):
                state = 5
            else:
                assert not re_32.match(line), _(
                    b'must have at least some revisions'
                )

        elif state == 5:
            # expecting revision number and possibly (ignored) lock indication
            # we create the logentry here from values stored in states 0 to 4,
            # as this state is re-entered for subsequent revisions of a file.
            match = re_50.match(line)
            assert match, _(b'expected revision number')
            e = logentry(
                rcs=scache(rcs),
                file=scache(filename),
                revision=tuple([int(x) for x in match.group(1).split(b'.')]),
                branches=[],
                parent=None,
                commitid=None,
                mergepoint=None,
                branchpoints=set(),
            )

            state = 6

        elif state == 6:
            # expecting date, author, state, lines changed
            match = re_60.match(line)
            assert match, _(b'revision must be followed by date line')
            d = match.group(1)
            if d[2] == b'/':
                # Y2K
                d = b'19' + d

            if len(d.split()) != 3:
                # cvs log dates always in GMT
                d = d + b' UTC'
            e.date = dateutil.parsedate(
                d,
                [
                    b'%y/%m/%d %H:%M:%S',
                    b'%Y/%m/%d %H:%M:%S',
                    b'%Y-%m-%d %H:%M:%S',
                ],
            )
            e.author = scache(match.group(2))
            e.dead = match.group(3).lower() == b'dead'

            if match.group(5):
                if match.group(6):
                    e.lines = (int(match.group(5)), int(match.group(6)))
                else:
                    e.lines = (int(match.group(5)), 0)
            elif match.group(6):
                e.lines = (0, int(match.group(6)))
            else:
                e.lines = None

            if match.group(7):  # cvs 1.12 commitid
                e.commitid = match.group(8)

            if match.group(9):  # cvsnt mergepoint
                myrev = match.group(10).split(b'.')
                if len(myrev) == 2:  # head
                    e.mergepoint = b'HEAD'
                else:
                    myrev = b'.'.join(myrev[:-2] + [b'0', myrev[-2]])
                    branches = [b for b in branchmap if branchmap[b] == myrev]
                    assert len(branches) == 1, (
                        b'unknown branch: %s' % e.mergepoint
                    )
                    e.mergepoint = branches[0]

            e.comment = []
            state = 7

        elif state == 7:
            # read the revision numbers of branches that start at this revision
            # or store the commit log message otherwise
            m = re_70.match(line)
            if m:
                e.branches = [
                    tuple([int(y) for y in x.strip().split(b'.')])
                    for x in m.group(1).split(b';')
                ]
                state = 8
            elif re_31.match(line) and re_50.match(peek):
                state = 5
                store = True
            elif re_32.match(line):
                state = 0
                store = True
            else:
                e.comment.append(line)

        elif state == 8:
            # store commit log message
            if re_31.match(line):
                cpeek = peek
                if cpeek.endswith(b'\n'):
                    cpeek = cpeek[:-1]
                if re_50.match(cpeek):
                    state = 5
                    store = True
                else:
                    e.comment.append(line)
            elif re_32.match(line):
                state = 0
                store = True
            else:
                e.comment.append(line)

        # When a file is added on a branch B1, CVS creates a synthetic
        # dead trunk revision 1.1 so that the branch has a root.
        # Likewise, if you merge such a file to a later branch B2 (one
        # that already existed when the file was added on B1), CVS
        # creates a synthetic dead revision 1.1.x.1 on B2.  Don't drop
        # these revisions now, but mark them synthetic so
        # createchangeset() can take care of them.
        if (
            store
            and e.dead
            and e.revision[-1] == 1
            and len(e.comment) == 1  # 1.1 or 1.1.x.1
            and file_added_re.match(e.comment[0])
        ):
            ui.debug(
                b'found synthetic revision in %s: %r\n' % (e.rcs, e.comment[0])
            )
            e.synthetic = True

        if store:
            # clean up the results and save in the log.
            store = False
            e.tags = sorted([scache(x) for x in tags.get(e.revision, [])])
            e.comment = scache(b'\n'.join(e.comment))

            revn = len(e.revision)
            if revn > 3 and (revn % 2) == 0:
                e.branch = tags.get(e.revision[:-1], [None])[0]
            else:
                e.branch = None

            # find the branches starting from this revision
            branchpoints = set()
            for branch, revision in branchmap.items():
                revparts = tuple([int(i) for i in revision.split(b'.')])
                if len(revparts) < 2:  # bad tags
                    continue
                if revparts[-2] == 0 and revparts[-1] % 2 == 0:
                    # normal branch
                    if revparts[:-2] == e.revision:
                        branchpoints.add(branch)
                elif revparts == (1, 1, 1):  # vendor branch
                    if revparts in e.branches:
                        branchpoints.add(branch)
            e.branchpoints = branchpoints

            log.append(e)

            rcsmap[e.rcs.replace(b'/Attic/', b'/')] = e.rcs

            if len(log) % 100 == 0:
                ui.status(
                    stringutil.ellipsis(b'%d %s' % (len(log), e.file), 80)
                    + b'\n'
                )

    log.sort(key=lambda x: (x.rcs, x.revision))

    # find parent revisions of individual files
    versions = {}
    for e in sorted(oldlog, key=lambda x: (x.rcs, x.revision)):
        rcs = e.rcs.replace(b'/Attic/', b'/')
        if rcs in rcsmap:
            e.rcs = rcsmap[rcs]
        branch = e.revision[:-1]
        versions[(e.rcs, branch)] = e.revision

    for e in log:
        branch = e.revision[:-1]
        p = versions.get((e.rcs, branch), None)
        if p is None:
            p = e.revision[:-2]
        e.parent = p
        versions[(e.rcs, branch)] = e.revision

    # update the log cache
    if cache:
        if log:
            # join up the old and new logs
            log.sort(key=lambda x: x.date)

            if oldlog and oldlog[-1].date >= log[0].date:
                raise logerror(
                    _(
                        b'log cache overlaps with new log entries,'
                        b' re-run without cache.'
                    )
                )

            log = oldlog + log

            # write the new cachefile
            ui.note(_(b'writing cvs log cache %s\n') % cachefile)

            with open(cachefile, 'wb') as fp:
                pickle.dump(log, fp)
        else:
            log = oldlog

    ui.status(_(b'%d log entries\n') % len(log))

    encodings = ui.configlist(b'convert', b'cvsps.logencoding')
    if encodings:

        def revstr(r):
            # this is needed, because logentry.revision is a tuple of "int"
            # (e.g. (1, 2) for "1.2")
            return b'.'.join(pycompat.maplist(pycompat.bytestr, r))

        for entry in log:
            comment = entry.comment
            for e in encodings:
                try:
                    entry.comment = comment.decode(pycompat.sysstr(e)).encode(
                        'utf-8'
                    )
                    if ui.debugflag:
                        ui.debug(
                            b"transcoding by %s: %s of %s\n"
                            % (e, revstr(entry.revision), entry.file)
                        )
                    break
                except UnicodeDecodeError:
                    pass  # try next encoding
                except LookupError as inst:  # unknown encoding, maybe
                    raise error.Abort(
                        pycompat.bytestr(inst),
                        hint=_(
                            b'check convert.cvsps.logencoding configuration'
                        ),
                    )
            else:
                raise error.Abort(
                    _(
                        b"no encoding can transcode"
                        b" CVS log message for %s of %s"
                    )
                    % (revstr(entry.revision), entry.file),
                    hint=_(b'check convert.cvsps.logencoding configuration'),
                )

    hook.hook(ui, None, b"cvslog", True, log=log)

    return log


class changeset:
    """Class changeset has the following attributes:
    .id        - integer identifying this changeset (list index)
    .author    - author name as CVS knows it
    .branch    - name of branch this changeset is on, or None
    .comment   - commit message
    .commitid  - CVS commitid or None
    .date      - the commit date as a (time,tz) tuple
    .entries   - list of logentry objects in this changeset
    .parents   - list of one or two parent changesets
    .tags      - list of tags on this changeset
    .synthetic - from synthetic revision "file ... added on branch ..."
    .mergepoint- the branch that has been merged from or None
    .branchpoints- the branches that start at the current entry or empty
    """

    def __init__(self, **entries):
        self.id = None
        self.synthetic = False
        self.__dict__.update(entries)

    def __repr__(self):
        items = (
            b"%s=%r" % (k, self.__dict__[k]) for k in sorted(self.__dict__)
        )
        return b"%s(%s)" % (type(self).__name__, b", ".join(items))


def createchangeset(ui, log, fuzz=60, mergefrom=None, mergeto=None):
    '''Convert log into changesets.'''

    ui.status(_(b'creating changesets\n'))

    # try to order commitids by date
    mindate = {}
    for e in log:
        if e.commitid:
            if e.commitid not in mindate:
                mindate[e.commitid] = e.date
            else:
                mindate[e.commitid] = min(e.date, mindate[e.commitid])

    # Merge changesets
    log.sort(
        key=lambda x: (
            mindate.get(x.commitid, (-1, 0)),
            x.commitid or b'',
            x.comment,
            x.author,
            x.branch or b'',
            x.date,
            x.branchpoints,
        )
    )

    changesets = []
    files = set()
    c = None
    for i, e in enumerate(log):
        # Check if log entry belongs to the current changeset or not.

        # Since CVS is file-centric, two different file revisions with
        # different branchpoints should be treated as belonging to two
        # different changesets (and the ordering is important and not
        # honoured by cvsps at this point).
        #
        # Consider the following case:
        # foo 1.1 branchpoints: [MYBRANCH]
        # bar 1.1 branchpoints: [MYBRANCH, MYBRANCH2]
        #
        # Here foo is part only of MYBRANCH, but not MYBRANCH2, e.g. a
        # later version of foo may be in MYBRANCH2, so foo should be the
        # first changeset and bar the next and MYBRANCH and MYBRANCH2
        # should both start off of the bar changeset. No provisions are
        # made to ensure that this is, in fact, what happens.
        if not (
            c
            and e.branchpoints == c.branchpoints
            and (  # cvs commitids
                (e.commitid is not None and e.commitid == c.commitid)
                or (  # no commitids, use fuzzy commit detection
                    (e.commitid is None or c.commitid is None)
                    and e.comment == c.comment
                    and e.author == c.author
                    and e.branch == c.branch
                    and (
                        (c.date[0] + c.date[1])
                        <= (e.date[0] + e.date[1])
                        <= (c.date[0] + c.date[1]) + fuzz
                    )
                    and e.file not in files
                )
            )
        ):
            c = changeset(
                comment=e.comment,
                author=e.author,
                branch=e.branch,
                date=e.date,
                entries=[],
                mergepoint=e.mergepoint,
                branchpoints=e.branchpoints,
                commitid=e.commitid,
            )
            changesets.append(c)

            files = set()
            if len(changesets) % 100 == 0:
                t = b'%d %s' % (
                    len(changesets),
                    pycompat.byterepr(e.comment)[2:-1],
                )
                ui.status(stringutil.ellipsis(t, 80) + b'\n')

        c.entries.append(e)
        files.add(e.file)
        c.date = e.date  # changeset date is date of latest commit in it

    # Mark synthetic changesets

    for c in changesets:
        # Synthetic revisions always get their own changeset, because
        # the log message includes the filename.  E.g. if you add file3
        # and file4 on a branch, you get four log entries and three
        # changesets:
        #   "File file3 was added on branch ..." (synthetic, 1 entry)
        #   "File file4 was added on branch ..." (synthetic, 1 entry)
        #   "Add file3 and file4 to fix ..."     (real, 2 entries)
        # Hence the check for 1 entry here.
        c.synthetic = len(c.entries) == 1 and c.entries[0].synthetic

    # Sort files in each changeset

    def entitycompare(l, r):
        """Mimic cvsps sorting order"""
        l = l.file.split(b'/')
        r = r.file.split(b'/')
        nl = len(l)
        nr = len(r)
        n = min(nl, nr)
        for i in range(n):
            if i + 1 == nl and nl < nr:
                return -1
            elif i + 1 == nr and nl > nr:
                return +1
            elif l[i] < r[i]:
                return -1
            elif l[i] > r[i]:
                return +1
        return 0

    for c in changesets:
        c.entries.sort(key=functools.cmp_to_key(entitycompare))

    # Sort changesets by date

    odd = set()

    def cscmp(l, r):
        d = sum(l.date) - sum(r.date)
        if d:
            return d

        # detect vendor branches and initial commits on a branch
        le = {}
        for e in l.entries:
            le[e.rcs] = e.revision
        re = {}
        for e in r.entries:
            re[e.rcs] = e.revision

        d = 0
        for e in l.entries:
            if re.get(e.rcs, None) == e.parent:
                assert not d
                d = 1
                break

        for e in r.entries:
            if le.get(e.rcs, None) == e.parent:
                if d:
                    odd.add((l, r))
                d = -1
                break
        # By this point, the changesets are sufficiently compared that
        # we don't really care about ordering. However, this leaves
        # some race conditions in the tests, so we compare on the
        # number of files modified, the files contained in each
        # changeset, and the branchpoints in the change to ensure test
        # output remains stable.

        # recommended replacement for cmp from
        # https://docs.python.org/3.0/whatsnew/3.0.html
        c = lambda x, y: (x > y) - (x < y)
        # Sort bigger changes first.
        if not d:
            d = c(len(l.entries), len(r.entries))
        # Try sorting by filename in the change.
        if not d:
            d = c([e.file for e in l.entries], [e.file for e in r.entries])
        # Try and put changes without a branch point before ones with
        # a branch point.
        if not d:
            d = c(len(l.branchpoints), len(r.branchpoints))
        return d

    changesets.sort(key=functools.cmp_to_key(cscmp))

    # Collect tags

    globaltags = {}
    for c in changesets:
        for e in c.entries:
            for tag in e.tags:
                # remember which is the latest changeset to have this tag
                globaltags[tag] = c

    for c in changesets:
        tags = set()
        for e in c.entries:
            tags.update(e.tags)
        # remember tags only if this is the latest changeset to have it
        c.tags = sorted(tag for tag in tags if globaltags[tag] is c)

    # Find parent changesets, handle {{mergetobranch BRANCHNAME}}
    # by inserting dummy changesets with two parents, and handle
    # {{mergefrombranch BRANCHNAME}} by setting two parents.

    if mergeto is None:
        mergeto = br'{{mergetobranch ([-\w]+)}}'
    if mergeto:
        mergeto = re.compile(mergeto)

    if mergefrom is None:
        mergefrom = br'{{mergefrombranch ([-\w]+)}}'
    if mergefrom:
        mergefrom = re.compile(mergefrom)

    versions = {}  # changeset index where we saw any particular file version
    branches = {}  # changeset index where we saw a branch
    n = len(changesets)
    i = 0
    while i < n:
        c = changesets[i]

        for f in c.entries:
            versions[(f.rcs, f.revision)] = i

        p = None
        if c.branch in branches:
            p = branches[c.branch]
        else:
            # first changeset on a new branch
            # the parent is a changeset with the branch in its
            # branchpoints such that it is the latest possible
            # commit without any intervening, unrelated commits.

            for candidate in range(i):
                if c.branch not in changesets[candidate].branchpoints:
                    if p is not None:
                        break
                    continue
                p = candidate

        c.parents = []
        if p is not None:
            p = changesets[p]

            # Ensure no changeset has a synthetic changeset as a parent.
            while p.synthetic:
                assert len(p.parents) <= 1, _(
                    b'synthetic changeset cannot have multiple parents'
                )
                if p.parents:
                    p = p.parents[0]
                else:
                    p = None
                    break

            if p is not None:
                c.parents.append(p)

        if c.mergepoint:
            if c.mergepoint == b'HEAD':
                c.mergepoint = None
            c.parents.append(changesets[branches[c.mergepoint]])

        if mergefrom:
            m = mergefrom.search(c.comment)
            if m:
                m = m.group(1)
                if m == b'HEAD':
                    m = None
                try:
                    candidate = changesets[branches[m]]
                except KeyError:
                    ui.warn(
                        _(
                            b"warning: CVS commit message references "
                            b"non-existent branch %r:\n%s\n"
                        )
                        % (pycompat.bytestr(m), c.comment)
                    )
                if m in branches and c.branch != m and not candidate.synthetic:
                    c.parents.append(candidate)

        if mergeto:
            m = mergeto.search(c.comment)
            if m:
                if m.groups():
                    m = m.group(1)
                    if m == b'HEAD':
                        m = None
                else:
                    m = None  # if no group found then merge to HEAD
                if m in branches and c.branch != m:
                    # insert empty changeset for merge
                    cc = changeset(
                        author=c.author,
                        branch=m,
                        date=c.date,
                        comment=b'convert-repo: CVS merge from branch %s'
                        % c.branch,
                        entries=[],
                        tags=[],
                        parents=[changesets[branches[m]], c],
                    )
                    changesets.insert(i + 1, cc)
                    branches[m] = i + 1

                    # adjust our loop counters now we have inserted a new entry
                    n += 1
                    i += 2
                    continue

        branches[c.branch] = i
        i += 1

    # Drop synthetic changesets (safe now that we have ensured no other
    # changesets can have them as parents).
    i = 0
    while i < len(changesets):
        if changesets[i].synthetic:
            del changesets[i]
        else:
            i += 1

    # Number changesets

    for i, c in enumerate(changesets):
        c.id = i + 1

    if odd:
        for l, r in odd:
            if l.id is not None and r.id is not None:
                ui.warn(
                    _(b'changeset %d is both before and after %d\n')
                    % (l.id, r.id)
                )

    ui.status(_(b'%d changeset entries\n') % len(changesets))

    hook.hook(ui, None, b"cvschangesets", True, changesets=changesets)

    return changesets


def debugcvsps(ui, *args, **opts):
    """Read CVS rlog for current directory or named path in
    repository, and convert the log to changesets based on matching
    commit log entries and dates.
    """
    opts = pycompat.byteskwargs(opts)
    if opts[b"new_cache"]:
        cache = b"write"
    elif opts[b"update_cache"]:
        cache = b"update"
    else:
        cache = None

    revisions = opts[b"revisions"]

    try:
        if args:
            log = []
            for d in args:
                log += createlog(ui, d, root=opts[b"root"], cache=cache)
        else:
            log = createlog(ui, root=opts[b"root"], cache=cache)
    except logerror as e:
        ui.write(b"%r\n" % e)
        return

    changesets = createchangeset(ui, log, opts[b"fuzz"])
    del log

    # Print changesets (optionally filtered)

    off = len(revisions)
    branches = {}  # latest version number in each branch
    ancestors = {}  # parent branch
    for cs in changesets:
        if opts[b"ancestors"]:
            if cs.branch not in branches and cs.parents and cs.parents[0].id:
                ancestors[cs.branch] = (
                    changesets[cs.parents[0].id - 1].branch,
                    cs.parents[0].id,
                )
            branches[cs.branch] = cs.id

        # limit by branches
        if (
            opts[b"branches"]
            and (cs.branch or b'HEAD') not in opts[b"branches"]
        ):
            continue

        if not off:
            # Note: trailing spaces on several lines here are needed to have
            #       bug-for-bug compatibility with cvsps.
            ui.write(b'---------------------\n')
            ui.writenoi18n(b'PatchSet %d \n' % cs.id)
            ui.writenoi18n(
                b'Date: %s\n'
                % dateutil.datestr(cs.date, b'%Y/%m/%d %H:%M:%S %1%2')
            )
            ui.writenoi18n(b'Author: %s\n' % cs.author)
            ui.writenoi18n(b'Branch: %s\n' % (cs.branch or b'HEAD'))
            ui.writenoi18n(
                b'Tag%s: %s \n'
                % (
                    [b'', b's'][len(cs.tags) > 1],
                    b','.join(cs.tags) or b'(none)',
                )
            )
            if cs.branchpoints:
                ui.writenoi18n(
                    b'Branchpoints: %s \n' % b', '.join(sorted(cs.branchpoints))
                )
            if opts[b"parents"] and cs.parents:
                if len(cs.parents) > 1:
                    ui.writenoi18n(
                        b'Parents: %s\n'
                        % (b','.join([(b"%d" % p.id) for p in cs.parents]))
                    )
                else:
                    ui.writenoi18n(b'Parent: %d\n' % cs.parents[0].id)

            if opts[b"ancestors"]:
                b = cs.branch
                r = []
                while b:
                    b, c = ancestors[b]
                    r.append(b'%s:%d:%d' % (b or b"HEAD", c, branches[b]))
                if r:
                    ui.writenoi18n(b'Ancestors: %s\n' % (b','.join(r)))

            ui.writenoi18n(b'Log:\n')
            ui.write(b'%s\n\n' % cs.comment)
            ui.writenoi18n(b'Members: \n')
            for f in cs.entries:
                fn = f.file
                if fn.startswith(opts[b"prefix"]):
                    fn = fn[len(opts[b"prefix"]) :]
                ui.write(
                    b'\t%s:%s->%s%s \n'
                    % (
                        fn,
                        b'.'.join([b"%d" % x for x in f.parent]) or b'INITIAL',
                        b'.'.join([(b"%d" % x) for x in f.revision]),
                        [b'', b'(DEAD)'][f.dead],
                    )
                )
            ui.write(b'\n')

        # have we seen the start tag?
        if revisions and off:
            if revisions[0] == (b"%d" % cs.id) or revisions[0] in cs.tags:
                off = False

        # see if we reached the end tag
        if len(revisions) > 1 and not off:
            if revisions[1] == (b"%d" % cs.id) or revisions[1] in cs.tags:
                break



# hook.py - hook support for mercurial
#
# Copyright 2007 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import contextlib
import errno
import os
import sys

from .i18n import _
from . import (
    demandimport,
    encoding,
    error,
    extensions,
    pycompat,
    util,
)
from .utils import (
    procutil,
    resourceutil,
    stringutil,
)


def pythonhook(ui, repo, htype, hname, funcname, args, throw):
    """call python hook. hook is callable object, looked up as
    name in python module. if callable returns "true", hook
    fails, else passes. if hook raises exception, treated as
    hook failure. exception propagates if throw is "true".

    reason for "true" meaning "hook failed" is so that
    unmodified commands (e.g. mercurial.commands.update) can
    be run as hooks without wrappers to convert return values."""

    if callable(funcname):
        obj = funcname
        funcname = obj.__module__ + "." + obj.__name__
    else:
        funcname = pycompat.sysstr(funcname)
        d = funcname.rfind('.')
        if d == -1:
            raise error.HookLoadError(
                _(b'%s hook is invalid: "%s" not in a module')
                % (hname, stringutil.forcebytestr(funcname))
            )
        modname = funcname[:d]
        oldpaths = sys.path
        if resourceutil.mainfrozen():
            # binary installs require sys.path manipulation
            modpath, modfile = os.path.split(modname)
            if modpath and modfile:
                sys.path = sys.path[:] + [modpath]
                modname = modfile
        with demandimport.deactivated():
            try:
                obj = __import__(pycompat.sysstr(modname))
            except (ImportError, SyntaxError):
                e1 = sys.exc_info()
                try:
                    # extensions are loaded with hgext_ prefix
                    obj = __import__("hgext_%s" % pycompat.sysstr(modname))
                except (ImportError, SyntaxError):
                    e2 = sys.exc_info()
                    if ui.tracebackflag:
                        ui.warn(
                            _(
                                b'exception from first failed import '
                                b'attempt:\n'
                            )
                        )
                    ui.traceback(e1)
                    if ui.tracebackflag:
                        ui.warn(
                            _(
                                b'exception from second failed import '
                                b'attempt:\n'
                            )
                        )
                    ui.traceback(e2)

                    if not ui.tracebackflag:
                        tracebackhint = _(
                            b'run with --traceback for stack trace'
                        )
                    else:
                        tracebackhint = None
                    msg = _(b'%s hook is invalid: import of "%s" failed')
                    msg %= (
                        stringutil.forcebytestr(hname),
                        stringutil.forcebytestr(modname),
                    )
                    raise error.HookLoadError(msg, hint=tracebackhint)
        sys.path = oldpaths
        try:
            for p in funcname.split('.')[1:]:
                obj = getattr(obj, p)
        except AttributeError:
            raise error.HookLoadError(
                _(b'%s hook is invalid: "%s" is not defined')
                % (hname, stringutil.forcebytestr(funcname))
            )
        if not callable(obj):
            raise error.HookLoadError(
                _(b'%s hook is invalid: "%s" is not callable')
                % (hname, stringutil.forcebytestr(funcname))
            )

    ui.note(
        _(b"calling hook %s: %s\n") % (hname, stringutil.forcebytestr(funcname))
    )
    starttime = util.timer()

    try:
        r = obj(ui=ui, repo=repo, hooktype=htype, **pycompat.strkwargs(args))
    except Exception as exc:
        if isinstance(exc, error.Abort):
            ui.warn(_(b'error: %s hook failed: %s\n') % (hname, exc.args[0]))
        else:
            ui.warn(
                _(b'error: %s hook raised an exception: %s\n')
                % (hname, stringutil.forcebytestr(exc))
            )
        if throw:
            raise
        if not ui.tracebackflag:
            ui.warn(_(b'(run with --traceback for stack trace)\n'))
        ui.traceback()
        return True, True
    finally:
        duration = util.timer() - starttime
        ui.log(
            b'pythonhook',
            b'pythonhook-%s: %s finished in %0.2f seconds\n',
            htype,
            stringutil.forcebytestr(funcname),
            duration,
        )
    if r:
        if throw:
            raise error.HookAbort(_(b'%s hook failed') % hname)
        ui.warn(_(b'warning: %s hook failed\n') % hname)
    return r, False


def _exthook(ui, repo, htype, name, cmd, args, throw):
    starttime = util.timer()
    env = {}

    # make in-memory changes visible to external process
    if repo is not None:
        tr = repo.currenttransaction()
        repo.dirstate.write(tr)
        if tr and tr.writepending():
            env[b'HG_PENDING'] = repo.root
    env[b'HG_HOOKTYPE'] = htype
    env[b'HG_HOOKNAME'] = name

    if ui.config(b'hooks', b'%s:run-with-plain' % name) == b'auto':
        plain = ui.plain()
    else:
        plain = ui.configbool(b'hooks', b'%s:run-with-plain' % name)
    if plain:
        env[b'HGPLAIN'] = b'1'
    else:
        env[b'HGPLAIN'] = b''

    for k, v in args.items():
        # transaction changes can accumulate MBs of data, so skip it
        # for external hooks
        if k == b'changes':
            continue
        if callable(v):
            v = v()
        if isinstance(v, (dict, list)):
            v = stringutil.pprint(v)
        env[b'HG_' + k.upper()] = v

    if ui.configbool(b'hooks', b'tonative.%s' % name, False):
        oldcmd = cmd
        cmd = procutil.shelltonative(cmd, env)
        if cmd != oldcmd:
            ui.note(_(b'converting hook "%s" to native\n') % name)

    ui.note(_(b"running hook %s: %s\n") % (name, cmd))

    if repo:
        cwd = repo.root
    else:
        cwd = encoding.getcwd()
    r = ui.system(cmd, environ=env, cwd=cwd, blockedtag=b'exthook-%s' % (name,))

    if repo is not None and repo.currentwlock() is None:
        repo.invalidatedirstate()
    if repo is not None and repo.currentlock() is None:
        repo.invalidate()

    duration = util.timer() - starttime
    ui.log(
        b'exthook',
        b'exthook-%s: %s finished in %0.2f seconds\n',
        name,
        cmd,
        duration,
    )
    if r:
        desc = procutil.explainexit(r)
        if throw:
            raise error.HookAbort(_(b'%s hook %s') % (name, desc))
        ui.warn(_(b'warning: %s hook %s\n') % (name, desc))
    return r


# represent an untrusted hook command
_fromuntrusted = object()


def _allhooks(ui):
    """return a list of (hook-id, cmd) pairs sorted by priority"""
    hooks = _hookitems(ui)
    # Be careful in this section, propagating the real commands from untrusted
    # sources would create a security vulnerability, make sure anything altered
    # in that section uses "_fromuntrusted" as its command.
    untrustedhooks = _hookitems(ui, _untrusted=True)
    for name, value in untrustedhooks.items():
        trustedvalue = hooks.get(name, ((), (), name, _fromuntrusted))
        if value != trustedvalue:
            (lp, lo, lk, lv) = trustedvalue
            hooks[name] = (lp, lo, lk, _fromuntrusted)
    # (end of the security sensitive section)
    return [(k, v) for p, o, k, v in sorted(hooks.values())]


def _hookitems(ui, _untrusted=False):
    """return all hooks items ready to be sorted"""
    hooks = {}
    for name, cmd in ui.configitems(b'hooks', untrusted=_untrusted):
        if (
            name.startswith(b'priority.')
            or name.startswith(b'tonative.')
            or b':' in name
        ):
            continue

        priority = ui.configint(b'hooks', b'priority.%s' % name, 0)
        hooks[name] = ((-priority,), (len(hooks),), name, cmd)
    return hooks


_redirect = False


def redirect(state):
    global _redirect
    _redirect = state


def hashook(ui, htype):
    """return True if a hook is configured for 'htype'"""
    if not ui.callhooks:
        return False
    for hname, cmd in _allhooks(ui):
        if hname.split(b'.')[0] == htype and cmd:
            return True
    return False


def hook(ui, repo, htype, throw=False, **args):
    if not ui.callhooks:
        return False

    hooks = []
    for hname, cmd in _allhooks(ui):
        if hname.split(b'.')[0] == htype and cmd:
            hooks.append((hname, cmd))

    res = runhooks(ui, repo, htype, hooks, throw=throw, **args)
    r = False
    for hname, cmd in hooks:
        r = res[hname][0] or r
    return r


@contextlib.contextmanager
def redirect_stdio():
    """Redirects stdout to stderr, if possible."""

    oldstdout = -1
    try:
        if _redirect:
            try:
                stdoutno = procutil.stdout.fileno()
                stderrno = procutil.stderr.fileno()
                # temporarily redirect stdout to stderr, if possible
                if stdoutno >= 0 and stderrno >= 0:
                    procutil.stdout.flush()
                    oldstdout = os.dup(stdoutno)
                    os.dup2(stderrno, stdoutno)
            except (OSError, AttributeError):
                # files seem to be bogus, give up on redirecting (WSGI, etc)
                pass

        yield

    finally:
        # The stderr is fully buffered on Windows when connected to a pipe.
        # A forcible flush is required to make small stderr data in the
        # remote side available to the client immediately.
        try:
            procutil.stderr.flush()
        except OSError as err:
            if err.errno not in (errno.EPIPE, errno.EIO, errno.EBADF):
                raise error.StdioError(err)

        if _redirect and oldstdout >= 0:
            try:
                procutil.stdout.flush()  # write hook output to stderr fd
            except OSError as err:
                if err.errno not in (errno.EPIPE, errno.EIO, errno.EBADF):
                    raise error.StdioError(err)
            os.dup2(oldstdout, stdoutno)
            os.close(oldstdout)


def runhooks(ui, repo, htype, hooks, throw=False, **args):
    args = pycompat.byteskwargs(args)
    res = {}

    with redirect_stdio():
        for hname, cmd in hooks:
            if cmd is _fromuntrusted:
                if throw:
                    raise error.HookAbort(
                        _(b'untrusted hook %s not executed') % hname,
                        hint=_(b"see 'hg help config.trusted'"),
                    )
                ui.warn(_(b'warning: untrusted hook %s not executed\n') % hname)
                r = 1
                raised = False
            elif callable(cmd):
                r, raised = pythonhook(ui, repo, htype, hname, cmd, args, throw)
            elif cmd.startswith(b'python:'):
                if cmd.count(b':') >= 2:
                    path, cmd = cmd[7:].rsplit(b':', 1)
                    path = util.expandpath(path)
                    if repo:
                        path = os.path.join(repo.root, path)
                    try:
                        mod_name = 'hghook.%s' % pycompat.sysstr(hname)
                        mod = extensions.loadpath(path, mod_name)
                    except Exception:
                        ui.write(_(b"loading %s hook failed:\n") % hname)
                        raise
                    hookfn = getattr(mod, pycompat.sysstr(cmd))
                else:
                    hookfn = cmd[7:].strip()
                r, raised = pythonhook(
                    ui, repo, htype, hname, hookfn, args, throw
                )
            else:
                r = _exthook(ui, repo, htype, hname, cmd, args, throw)
                raised = False

            res[hname] = r, raised

    return res



# split.py - split a changeset into smaller ones
#
# Copyright 2015 Laurent Charignon <lcharignon@fb.com>
# Copyright 2017 Facebook, Inc.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
"""command to split a changeset into smaller ones (EXPERIMENTAL)"""

from __future__ import annotations

from mercurial.i18n import _

from mercurial.node import (
    nullrev,
    short,
)

from mercurial import (
    bookmarks,
    cmdutil,
    commands,
    error,
    hg,
    logcmdutil,
    registrar,
    revsetlang,
    rewriteutil,
    scmutil,
    util,
)

# allow people to use split without explicitly enabling rebase extension
from . import rebase

cmdtable = {}
command = registrar.command(cmdtable)

# Note for extension authors: ONLY specify testedwith = 'ships-with-hg-core' for
# extensions which SHIP WITH MERCURIAL. Non-mainline extensions should
# be specifying the version(s) of Mercurial they are tested with, or
# leave the attribute unspecified.
testedwith = b'ships-with-hg-core'


@command(
    b'split',
    [
        (b'r', b'rev', b'', _(b"revision to split"), _(b'REV')),
        (b'', b'rebase', True, _(b'rebase descendants after split')),
    ]
    + cmdutil.commitopts2,
    _(b'hg split [--no-rebase] [[-r] REV]'),
    helpcategory=command.CATEGORY_CHANGE_MANAGEMENT,
    helpbasic=True,
)
def split(ui, repo, *revs, **opts):
    """split a changeset into smaller ones

    Repeatedly prompt changes and commit message for new changesets until there
    is nothing left in the original changeset.

    If --rev was not given, split the working directory parent.

    By default, rebase connected non-obsoleted descendants onto the new
    changeset. Use --no-rebase to avoid the rebase.
    """
    revlist = []
    if opts.get('rev'):
        revlist.append(opts.get('rev'))
    revlist.extend(revs)
    with repo.wlock(), repo.lock():
        tr = repo.transaction(b'split')
        # If the rebase somehow runs into conflicts, make sure
        # we close the transaction so the user can continue it.
        with util.acceptintervention(tr):
            revs = logcmdutil.revrange(repo, revlist or [b'.'])
            if len(revs) > 1:
                raise error.InputError(_(b'cannot split multiple revisions'))

            rev = revs.first()
            # Handle nullrev specially here (instead of leaving for precheck()
            # below) so we get a nicer message and error code.
            if rev is None or rev == nullrev:
                ui.status(_(b'nothing to split\n'))
                return 1
            ctx = repo[rev]
            if ctx.node() is None:
                raise error.InputError(_(b'cannot split working directory'))

            if opts.get('rebase'):
                # Skip obsoleted descendants and their descendants so the rebase
                # won't cause conflicts for sure.
                descendants = list(repo.revs(b'(%d::) - (%d)', rev, rev))
                torebase = list(
                    repo.revs(
                        b'%ld - (%ld & obsolete())::', descendants, descendants
                    )
                )
            else:
                torebase = []
            rewriteutil.precheck(repo, [rev] + torebase, b'split')

            if len(ctx.parents()) > 1:
                raise error.InputError(_(b'cannot split a merge changeset'))

            cmdutil.bailifchanged(repo)

            # Deactivate bookmark temporarily so it won't get moved
            # unintentionally
            bname = repo._activebookmark
            if bname and repo._bookmarks[bname] != ctx.node():
                bookmarks.deactivate(repo)

            wnode = repo[b'.'].node()
            top = None
            try:
                top = dosplit(ui, repo, tr, ctx, **opts)
            finally:
                # top is None: split failed, need update --clean recovery.
                # wnode == ctx.node(): wnode split, no need to update.
                if top is None or wnode != ctx.node():
                    hg.clean(repo, wnode, show_stats=False)
                if bname:
                    bookmarks.activate(repo, bname)
            if torebase and top:
                dorebase(ui, repo, torebase, top)


def dosplit(ui, repo, tr, ctx, **opts):
    committed = []  # [ctx]

    # Set working parent to ctx.p1(), and keep working copy as ctx's content
    if ctx.node() != repo.dirstate.p1():
        hg.clean(repo, ctx.node(), show_stats=False)
    with repo.dirstate.changing_parents(repo):
        scmutil.movedirstate(repo, ctx.p1())

    # Any modified, added, removed, deleted result means split is incomplete
    def incomplete(repo):
        st = repo.status()
        return any((st.modified, st.added, st.removed, st.deleted))

    # Main split loop
    while incomplete(repo):
        if committed:
            header = _(
                b'HG: Splitting %s. So far it has been split into:\n'
            ) % short(ctx.node())
            # We don't want color codes in the commit message template, so
            # disable the label() template function while we render it.
            with ui.configoverride(
                {(b'templatealias', b'label(l,x)'): b"x"}, b'split'
            ):
                for c in committed:
                    summary = cmdutil.format_changeset_summary(ui, c, b'split')
                    header += _(b'HG: - %s\n') % summary
            header += _(
                b'HG: Write commit message for the next split changeset.\n'
            )
        else:
            header = _(
                b'HG: Splitting %s. Write commit message for the '
                b'first split changeset.\n'
            ) % short(ctx.node())
        opts.update(
            {
                'edit': True,
                'interactive': True,
                'message': header + ctx.description(),
            }
        )
        origctx = repo[b'.']
        commands.commit(ui, repo, **opts)
        newctx = repo[b'.']
        # Ensure user didn't do a "no-op" split (such as deselecting
        # everything).
        if origctx.node() != newctx.node():
            committed.append(newctx)

    if not committed:
        raise error.InputError(_(b'cannot split an empty revision'))

    if len(committed) != 1 or committed[0].node() != ctx.node():
        # Ensure we don't strip a node if we produce the same commit as already
        # exists
        scmutil.cleanupnodes(
            repo,
            {ctx.node(): [c.node() for c in committed]},
            operation=b'split',
            fixphase=True,
        )

    return committed[-1]


def dorebase(ui, repo, src, destctx):
    rebase.rebase(
        ui,
        repo,
        rev=[revsetlang.formatspec(b'%ld', src)],
        dest=revsetlang.formatspec(b'%d', destctx.rev()),
    )



# dagparser.py - parser and generator for concise description of DAGs
#
# Copyright 2010 Peter Arrenbrecht <peter@arrenbrecht.ch>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import re
import string

from .i18n import _
from . import (
    error,
    pycompat,
)
from .utils import stringutil


def parsedag(desc):
    '''parses a DAG from a concise textual description; generates events

    "+n" is a linear run of n nodes based on the current default parent
    "." is a single node based on the current default parent
    "$" resets the default parent to -1 (implied at the start);
        otherwise the default parent is always the last node created
    "<p" sets the default parent to the backref p
    "*p" is a fork at parent p, where p is a backref
    "*p1/p2/.../pn" is a merge of parents p1..pn, where the pi are backrefs
    "/p2/.../pn" is a merge of the preceding node and p2..pn
    ":name" defines a label for the preceding node; labels can be redefined
    "@text" emits an annotation event for text
    "!command" emits an action event for the current node
    "!!my command\n" is like "!", but to the end of the line
    "#...\n" is a comment up to the end of the line

    Whitespace between the above elements is ignored.

    A backref is either
     * a number n, which references the node curr-n, where curr is the current
       node, or
     * the name of a label you placed earlier using ":name", or
     * empty to denote the default parent.

    All string valued-elements are either strictly alphanumeric, or must
    be enclosed in double quotes ("..."), with "\" as escape character.

    Generates sequence of

      ('n', (id, [parentids])) for node creation
      ('l', (id, labelname)) for labels on nodes
      ('a', text) for annotations
      ('c', command) for actions (!)
      ('C', command) for line actions (!!)

    Examples
    --------

    Example of a complex graph (output not shown for brevity):

        >>> len(list(parsedag(b"""
        ...
        ... +3         # 3 nodes in linear run
        ... :forkhere  # a label for the last of the 3 nodes from above
        ... +5         # 5 more nodes on one branch
        ... :mergethis # label again
        ... <forkhere  # set default parent to labeled fork node
        ... +10        # 10 more nodes on a parallel branch
        ... @stable    # following nodes will be annotated as "stable"
        ... +5         # 5 nodes in stable
        ... !addfile   # custom command; could trigger new file in next node
        ... +2         # two more nodes
        ... /mergethis # merge last node with labeled node
        ... +4         # 4 more nodes descending from merge node
        ...
        ... """)))
        34

    Empty list:

        >>> list(parsedag(b""))
        []

    A simple linear run:

        >>> list(parsedag(b"+3"))
        [('n', (0, [-1])), ('n', (1, [0])), ('n', (2, [1]))]

    Some non-standard ways to define such runs:

        >>> list(parsedag(b"+1+2"))
        [('n', (0, [-1])), ('n', (1, [0])), ('n', (2, [1]))]

        >>> list(parsedag(b"+1*1*"))
        [('n', (0, [-1])), ('n', (1, [0])), ('n', (2, [1]))]

        >>> list(parsedag(b"*"))
        [('n', (0, [-1]))]

        >>> list(parsedag(b"..."))
        [('n', (0, [-1])), ('n', (1, [0])), ('n', (2, [1]))]

    A fork and a join, using numeric back references:

        >>> list(parsedag(b"+2*2*/2"))
        [('n', (0, [-1])), ('n', (1, [0])), ('n', (2, [0])), ('n', (3, [2, 1]))]

        >>> list(parsedag(b"+2<2+1/2"))
        [('n', (0, [-1])), ('n', (1, [0])), ('n', (2, [0])), ('n', (3, [2, 1]))]

    Placing a label:

        >>> list(parsedag(b"+1 :mylabel +1"))
        [('n', (0, [-1])), ('l', (0, 'mylabel')), ('n', (1, [0]))]

    An empty label (silly, really):

        >>> list(parsedag(b"+1:+1"))
        [('n', (0, [-1])), ('l', (0, '')), ('n', (1, [0]))]

    Fork and join, but with labels instead of numeric back references:

        >>> list(parsedag(b"+1:f +1:p2 *f */p2"))
        [('n', (0, [-1])), ('l', (0, 'f')), ('n', (1, [0])), ('l', (1, 'p2')),
         ('n', (2, [0])), ('n', (3, [2, 1]))]

        >>> list(parsedag(b"+1:f +1:p2 <f +1 /p2"))
        [('n', (0, [-1])), ('l', (0, 'f')), ('n', (1, [0])), ('l', (1, 'p2')),
         ('n', (2, [0])), ('n', (3, [2, 1]))]

    Restarting from the root:

        >>> list(parsedag(b"+1 $ +1"))
        [('n', (0, [-1])), ('n', (1, [-1]))]

    Annotations, which are meant to introduce sticky state for subsequent nodes:

        >>> list(parsedag(b"+1 @ann +1"))
        [('n', (0, [-1])), ('a', 'ann'), ('n', (1, [0]))]

        >>> list(parsedag(b'+1 @"my annotation" +1'))
        [('n', (0, [-1])), ('a', 'my annotation'), ('n', (1, [0]))]

    Commands, which are meant to operate on the most recently created node:

        >>> list(parsedag(b"+1 !cmd +1"))
        [('n', (0, [-1])), ('c', 'cmd'), ('n', (1, [0]))]

        >>> list(parsedag(b'+1 !"my command" +1'))
        [('n', (0, [-1])), ('c', 'my command'), ('n', (1, [0]))]

        >>> list(parsedag(b'+1 !!my command line\\n +1'))
        [('n', (0, [-1])), ('C', 'my command line'), ('n', (1, [0]))]

    Comments, which extend to the end of the line:

        >>> list(parsedag(b'+1 # comment\\n+1'))
        [('n', (0, [-1])), ('n', (1, [0]))]

    Error:

        >>> try: list(parsedag(b'+1 bad'))
        ... except Exception as e: print(pycompat.sysstr(bytes(e)))
        invalid character in dag description: bad...

    '''
    if not desc:
        return

    # pytype: disable=wrong-arg-types
    wordchars = pycompat.bytestr(string.ascii_letters + string.digits)
    # pytype: enable=wrong-arg-types

    labels = {}
    p1 = -1
    r = 0

    def resolve(ref):
        if not ref:
            return p1
        # pytype: disable=wrong-arg-types
        elif ref[0] in pycompat.bytestr(string.digits):
            # pytype: enable=wrong-arg-types
            return r - int(ref)
        else:
            return labels[ref]

    chiter = pycompat.iterbytestr(desc)

    def nextch():
        return next(chiter, b'\0')

    def nextrun(c, allow):
        s = b''
        while c in allow:
            s += c
            c = nextch()
        return c, s

    def nextdelimited(c, limit, escape):
        s = b''
        while c != limit:
            if c == escape:
                c = nextch()
            s += c
            c = nextch()
        return nextch(), s

    def nextstring(c):
        if c == b'"':
            return nextdelimited(nextch(), b'"', b'\\')
        else:
            return nextrun(c, wordchars)

    c = nextch()
    while c != b'\0':
        # pytype: disable=wrong-arg-types
        while c in pycompat.bytestr(string.whitespace):
            # pytype: enable=wrong-arg-types
            c = nextch()
        if c == b'.':
            yield b'n', (r, [p1])
            p1 = r
            r += 1
            c = nextch()
        elif c == b'+':
            # pytype: disable=wrong-arg-types
            c, digs = nextrun(nextch(), pycompat.bytestr(string.digits))
            # pytype: enable=wrong-arg-types
            n = int(digs)
            for i in range(0, n):
                yield b'n', (r, [p1])
                p1 = r
                r += 1
        elif c in b'*/':
            if c == b'*':
                c = nextch()
            c, pref = nextstring(c)
            prefs = [pref]
            while c == b'/':
                c, pref = nextstring(nextch())
                prefs.append(pref)
            ps = [resolve(ref) for ref in prefs]
            yield b'n', (r, ps)
            p1 = r
            r += 1
        elif c == b'<':
            c, ref = nextstring(nextch())
            p1 = resolve(ref)
        elif c == b':':
            c, name = nextstring(nextch())
            labels[name] = p1
            yield b'l', (p1, name)
        elif c == b'@':
            c, text = nextstring(nextch())
            yield b'a', text
        elif c == b'!':
            c = nextch()
            if c == b'!':
                cmd = b''
                c = nextch()
                while c not in b'\n\r\0':
                    cmd += c
                    c = nextch()
                yield b'C', cmd
            else:
                c, cmd = nextstring(c)
                yield b'c', cmd
        elif c == b'#':
            while c not in b'\n\r\0':
                c = nextch()
        elif c == b'$':
            p1 = -1
            c = nextch()
        elif c == b'\0':
            return  # in case it was preceded by whitespace
        else:
            s = b''
            i = 0
            while c != b'\0' and i < 10:
                s += c
                i += 1
                c = nextch()
            raise error.Abort(
                _(b'invalid character in dag description: %s...') % s
            )


def dagtextlines(
    events,
    addspaces=True,
    wraplabels=False,
    wrapannotations=False,
    wrapcommands=False,
    wrapnonlinear=False,
    usedots=False,
    maxlinewidth=70,
):
    '''generates single lines for dagtext()'''

    def wrapstring(text):
        if re.match(b"^[0-9a-z]*$", text):
            return text
        return b'"' + text.replace(b'\\', b'\\\\').replace(b'"', b'\"') + b'"'

    def gen():
        labels = {}
        run = 0
        wantr = 0
        needroot = False
        for kind, data in events:
            if kind == b'n':
                r, ps = data

                # sanity check
                if r != wantr:
                    raise error.Abort(_(b"expected id %i, got %i") % (wantr, r))
                if not ps:
                    ps = [-1]
                else:
                    for p in ps:
                        if p >= r:
                            raise error.Abort(
                                _(
                                    b"parent id %i is larger than "
                                    b"current id %i"
                                )
                                % (p, r)
                            )
                wantr += 1

                # new root?
                p1 = r - 1
                if len(ps) == 1 and ps[0] == -1:
                    if needroot:
                        if run:
                            yield b'+%d' % run
                            run = 0
                        if wrapnonlinear:
                            yield b'\n'
                        yield b'$'
                        p1 = -1
                    else:
                        needroot = True
                if len(ps) == 1 and ps[0] == p1:
                    if usedots:
                        yield b"."
                    else:
                        run += 1
                else:
                    if run:
                        yield b'+%d' % run
                        run = 0
                    if wrapnonlinear:
                        yield b'\n'
                    prefs = []
                    for p in ps:
                        if p == p1:
                            prefs.append(b'')
                        elif p in labels:
                            prefs.append(labels[p])
                        else:
                            prefs.append(b'%d' % (r - p))
                    yield b'*' + b'/'.join(prefs)
            else:
                if run:
                    yield b'+%d' % run
                    run = 0
                if kind == b'l':
                    rid, name = data
                    labels[rid] = name
                    yield b':' + name
                    if wraplabels:
                        yield b'\n'
                elif kind == b'c':
                    yield b'!' + wrapstring(data)
                    if wrapcommands:
                        yield b'\n'
                elif kind == b'C':
                    yield b'!!' + data
                    yield b'\n'
                elif kind == b'a':
                    if wrapannotations:
                        yield b'\n'
                    yield b'@' + wrapstring(data)
                elif kind == b'#':
                    yield b'#' + data
                    yield b'\n'
                else:
                    raise error.Abort(
                        _(b"invalid event type in dag: ('%s', '%s')")
                        % (
                            stringutil.escapestr(kind),
                            stringutil.escapestr(data),
                        )
                    )
        if run:
            yield b'+%d' % run

    line = b''
    for part in gen():
        if part == b'\n':
            if line:
                yield line
                line = b''
        else:
            if len(line) + len(part) >= maxlinewidth:
                yield line
                line = b''
            elif addspaces and line and part != b'.':
                line += b' '
            line += part
    if line:
        yield line


def dagtext(
    dag,
    addspaces=True,
    wraplabels=False,
    wrapannotations=False,
    wrapcommands=False,
    wrapnonlinear=False,
    usedots=False,
    maxlinewidth=70,
):
    """generates lines of a textual representation for a dag event stream

    events should generate what parsedag() does, so:

      ('n', (id, [parentids])) for node creation
      ('l', (id, labelname)) for labels on nodes
      ('a', text) for annotations
      ('c', text) for commands
      ('C', text) for line commands ('!!')
      ('#', text) for comment lines

    Parent nodes must come before child nodes.

    Examples
    --------

    Linear run:

        >>> dagtext([(b'n', (0, [-1])), (b'n', (1, [0]))])
        '+2'

    Two roots:

        >>> dagtext([(b'n', (0, [-1])), (b'n', (1, [-1]))])
        '+1 $ +1'

    Fork and join:

        >>> dagtext([(b'n', (0, [-1])), (b'n', (1, [0])), (b'n', (2, [0])),
        ...          (b'n', (3, [2, 1]))])
        '+2 *2 */2'

    Fork and join with labels:

        >>> dagtext([(b'n', (0, [-1])), (b'l', (0, b'f')), (b'n', (1, [0])),
        ...          (b'l', (1, b'p2')), (b'n', (2, [0])), (b'n', (3, [2, 1]))])
        '+1 :f +1 :p2 *f */p2'

    Annotations:

        >>> dagtext([(b'n', (0, [-1])), (b'a', b'ann'), (b'n', (1, [0]))])
        '+1 @ann +1'

        >>> dagtext([(b'n', (0, [-1])),
        ...          (b'a', b'my annotation'),
        ...          (b'n', (1, [0]))])
        '+1 @"my annotation" +1'

    Commands:

        >>> dagtext([(b'n', (0, [-1])), (b'c', b'cmd'), (b'n', (1, [0]))])
        '+1 !cmd +1'

        >>> dagtext([(b'n', (0, [-1])),
        ...          (b'c', b'my command'),
        ...          (b'n', (1, [0]))])
        '+1 !"my command" +1'

        >>> dagtext([(b'n', (0, [-1])),
        ...          (b'C', b'my command line'),
        ...          (b'n', (1, [0]))])
        '+1 !!my command line\\n+1'

    Comments:

        >>> dagtext([(b'n', (0, [-1])), (b'#', b' comment'), (b'n', (1, [0]))])
        '+1 # comment\\n+1'

        >>> dagtext([])
        ''

    Combining parsedag and dagtext:

        >>> dagtext(parsedag(b'+1 :f +1 :p2 *f */p2'))
        '+1 :f +1 :p2 *f */p2'

    """
    return b"\n".join(
        dagtextlines(
            dag,
            addspaces,
            wraplabels,
            wrapannotations,
            wrapcommands,
            wrapnonlinear,
            usedots,
            maxlinewidth,
        )
    )



"""A lil' TOML parser."""

__all__ = ("loads", "load", "TOMLDecodeError")
__version__ = "1.2.3"  # DO NOT EDIT THIS LINE MANUALLY. LET bump2version UTILITY DO IT

from ._parser import TOMLDecodeError, load, loads

# Pretend this exception was created here.
TOMLDecodeError.__module__ = "tomli"
